# -*- coding: utf-8 -*-
"""Plots."""
# Ville Bergholm 2011-2014

from __future__ import division, absolute_import, print_function, unicode_literals

import numpy as np
from numpy import array, zeros, ones, sin, cos, tanh, sort, pi, r_, c_, linspace, outer
from numpy.linalg import eigvalsh
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

from .state import *
from .utils import copy_memoize, eighsort

__all__ = ['adiabatic_evolution', 'state_trajectory', 'bloch_sphere', 'correlation_simplex', 'pcolor',
           'asongoficeandfire', 'sphere']


def sphere(N=15):
    """X, Y, Z coordinate meshes for a unit sphere."""
    theta = linspace(0, pi, N)
    phi = linspace(0, 2*pi, 2*N)
    X = outer(sin(theta), cos(phi))
    Y = outer(sin(theta), sin(phi))
    Z = outer(cos(theta), ones(phi.shape))
    return X, Y, Z


def adiabatic_evolution(t, st, H_func, n=4):
    """Adiabatic evolution plot.

    Input: vector t of time instances, cell vector st of states corresponding
    to the times and time-dependant Hamiltonian function handle H_func.

    Plots the energies of the eigenstates of H_func(t(k)) as a function of t(k),
    and the overlap of st{k} with the n lowest final Hamiltonian eigenstates. 
    Useful for illustrating adiabatic evolution.
    """
    # Jacob D. Biamonte 2008
    # Ville Bergholm 2009-2010

    T = t[-1]  # final time
    H = H_func(T)

    n = min(n, H.shape[0])
    m = len(t)

    # find the n lowest eigenstates of the final Hamiltonian
    #d, v = scipy.sparse.linalg.eigs(H, n, which = 'SR')
    #ind = d.argsort()  # increasing real part
    d, v = eighsort(H)
    lowest = []
    for j in range(n):
        #j = ind[j]
        lowest.append(state(v[:, -j-1]))
    # TODO with degenerate states these are more or less random linear combinations of the basis states... overlaps are not meaningful

    energies = zeros((m, H.shape[0]))
    overlaps = zeros((m, n))
    for k in range(m):
        tt = t[k]
        H = H_func(tt)
        energies[k, :] = sort(eigvalsh(H).real)
        for j in range(n):
            overlaps[k, j] = lowest[j].fidelity(st[k]) ** 2 # squared overlap with lowest final states

    plt.subplot(2, 1, 1)
    plt.plot(t/T, energies)
    plt.grid(True)
    plt.title('Energy spectrum')
    plt.xlabel('Adiabatic time')
    plt.ylabel('Energy')
    plt.axis([0, 1, np.min(energies), np.max(energies)])


    plt.subplot(2, 1, 2)
    plt.plot(t/T, overlaps) #, 'LineWidth', 1.7)
    plt.grid(True)
    plt.title('Squared overlap of current state and final eigenstates')
    plt.xlabel('Adiabatic time')
    plt.ylabel('Probability')
    temp = []
    for k in range(n):
        temp.append('$|{0}\\rangle$'.format(k))
    plt.legend(temp)
    plt.axis([0, 1, 0, 1])
    # axis([0, 1, 0, max(overlaps)])


def bloch_sphere(ax=None):
    """Bloch sphere plot.

    Plots a Bloch sphere, a geometrical representation of the state space of a single qubit.
    Pure states are on the surface of the sphere, nonpure states inside it.
    The states |0> and |1> lie on the north and south poles of the sphere, respectively.

    s is a two dimensional state to be plotted.
    """
    # Ville Bergholm  2005-2012
    # James Whitfield 2010

    if ax == None:
        ax = plt.subplot(111, projection='3d')

    ax.hold(True)
    # surface
    X, Y, Z = sphere()
    ax.plot_surface(X, Y, Z, rstride = 1, cstride = 1, color = 'g', alpha = 0.2, linewidth = 0) #cmap = xxx
    ax.axis('tight')
    # poles
    coords = array([[0, 0, 1], [0, 0, -1]]).transpose()  # easier to read this way
    ax.scatter(*coords, c = 'r', marker = 'o')
    ax.text(0, 0,  1.1, '$|0\\rangle$')
    ax.text(0, 0, -1.2, '$|1\\rangle$')

    # TODO equator?
    #phi = linspace(0, 2*pi, 40);
    #plot3(cos(phi), sin(phi), zeros(size(phi)), 'k-');
    # labels
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    plt.show()
    return ax


def correlation_simplex(ax=None, labels='diagonal'):
    """Plots the correlations simplexes for two-qubit states.

    Plots the geometrical representation of the set of allowed
    correlations in a two-qubit state. For each group of three
    correlation variables the set is a tetrahedron.

    The groups are 'diagonal', 'pos' and 'neg'.
    For diagonal correlations the vertices correspond to the four Bell states.

    Returns the Axes instance and a vector of three linear
    indices denoting the correlations to be plotted as the x, y and z coordinates.

    NOTE the strange logic in the ordering of the pos and neg
    correlations follows the logic of the Bell state labeling convention, kind of.
    """
    # Ville Bergholm 2011-2012

    import mpl_toolkits.mplot3d as mplot3

    if ax == None:
        ax = plt.subplot(111, projection='3d')

    ax.hold(True)
    ax.grid(True)
    ax.view_init(20, -105)

    # tetrahedron
    # vertices and faces
    v = array([[-1, -1, -1], [-1, 1, 1], [1, -1, 1], [1, 1, -1]])
    f = [[0, 1, 2], [0, 2, 3], [0, 3, 1], [3, 2, 1]]
    polys = [v[k, :] for k in f]
    polyc = mplot3.art3d.Poly3DCollection(polys, color = 'g', alpha = 0.2)
    ax.add_collection(polyc)
    ax.set_xlim3d(-1, 1)
    ax.set_ylim3d(-1, 1)
    ax.set_zlim3d(-1, 1)

    # mark vertices
    ax.scatter([0], [0], [0], c = 'r', marker = '.')  # center
    ax.scatter(v[:, 0], v[:, 1], v[:, 2], c = 'r', marker = '.')  # vertices

    # label axes and vertices
    if labels == 'diagonal':
        ax.set_title('diagonal correlations')
        ax.set_xlabel('XX')
        ax.set_ylabel('YY')
        ax.set_zlabel('ZZ')
        ax.text(1.1, 1.1, -1.1, r'$|\Psi^+\rangle$')
        ax.text(1.1, -1.1, 1.1, r'$|\Phi^+\rangle$')
        ax.text(-1.1, 1.1, 1.1, r'$|\Phi^-\rangle$')
        ax.text(-1.2, -1.2, -1.2, r'$|\Psi^-\rangle$')
        ind = [5, 10, 15]

    elif labels == 'pos':
        ax.set_title('pos correlations')
        ax.set_xlabel('ZX')
        ax.set_ylabel('XY')
        ax.set_zlabel('YZ')
        ax.text(1.1, -1.1, 1.1, r'$|y+,0\rangle +|y-,1\rangle$')
        ax.text(-1.1, 1.1, 1.1, r'$|y+,0\rangle -|y-,1\rangle$')
        ax.text(1.1, 1.1, -1.1, r'$|y-,0\rangle +|y+,1\rangle$')
        ax.text(-1.2, -1.2, -1.2, r'$|y-,0\rangle -|y+,1\rangle$')
        ind = [7, 9, 14]

    elif labels == 'neg':
        ax.set_title('neg correlations')
        ax.set_xlabel('XZ')
        ax.set_ylabel('YX')
        ax.set_zlabel('ZY')
        ax.text(1.1, 1.1, -1.1, r'$|0,y-\rangle +|1,y+\rangle$')
        ax.text(-1.1, 1.1, 1.1, r'$|0,y+\rangle -|1,y-\rangle$')
        ax.text(1.1, -1.1, 1.1, r'$|0,y+\rangle +|1,y-\rangle$')
        ax.text(-1.2, -1.2, -1.2, r'$|0,y-\rangle -|1,y+\rangle$')
        ind = [13, 6, 11]

    elif labels == 'none':
        ind = []

    else:
        raise ValueError('Unknown set of correlations.')

    plt.show()
    return ax, ind


def state_trajectory(traj, reset=True, ax=None, color='b'):
    """Plot a state trajectory in the correlation representation.

    For a single-qubit system, plots the trajectory in the Bloch sphere.

    For a two-qubit system, plots the reduced single-qubit states (in
    Bloch spheres), as well as the interqubit correlations.

    traj is a list of generalized Bloch vectors.
    It can be obtained e.g. by using one of the continuous-time
    state propagation functions and feeding the results to
    bloch_vector.

    If reset is false, adds another trajectory to current plot
    without erasing it.

    Example 1: trajectory of s under the Hamiltonian H
      out = propagate(s, H, t, @(s,H) bloch_vector(s))
      bloch_trajectory(out)

    Example 2: just a single state s
      bloch_trajectory({bloch_vector(s)})
    """
    # Ville Bergholm  2006-2012

    if ax == None:
        ax = plt.subplot(111, projection='3d')

    def plot_traj(ax, A, ind):
        """Plots the trajectory formed by the correlations given in ind."""

        ax.scatter(A[0,  ind[0]],  A[0,  ind[1]],  A[0,  ind[2]], c = color, marker = 'x')
        # if we only have a single point, do not bother with these
        if len(A) > 1:
            ax.plot(A[:,  ind[0]],  A[:,  ind[1]],  A[:,  ind[2]], c = color)
            ax.scatter(A[-1, ind[0]],  A[-1, ind[1]],  A[-1, ind[2]], c = color, marker = 'o')


    if isinstance(traj, list):
        d = traj[0].size
    else:
        d = traj.size
    A = array(traj).reshape((-1, d), order='F')  # list index becomes the first dimension

    if len(A[0]) == 4:
        # single qubit
        if reset:
            bloch_sphere(ax)
        plot_traj(ax, A, [1, 2, 3])
  
    elif len(A[0]) == 16:
        # two qubits (or a single ququat...)

        # TODO split ax into subplots...
        fig = plt.gcf()
        if reset:
            gs = GridSpec(2, 3)

            ax = fig.add_subplot(gs[0, 0], projection = '3d')
            bloch_sphere(ax)
            ax.set_title('qubit A')
    
            ax = fig.add_subplot(gs[0, 1], projection = '3d')
            bloch_sphere(ax)
            ax.set_title('qubit B')

            ax = fig.add_subplot(gs[1, 0], projection = '3d')
            correlation_simplex(ax, labels = 'diagonal')

            ax = fig.add_subplot(gs[1, 1], projection = '3d')
            correlation_simplex(ax, labels = 'pos')

            ax = fig.add_subplot(gs[1, 2], projection = '3d')
            correlation_simplex(ax, labels = 'neg')

        # update existing axes instances
        qqq = fig.get_axes()
        plot_traj(qqq[0], A, [1, 2, 3])
        plot_traj(qqq[1], A, [4, 8, 12])
        plot_traj(qqq[2], A, [5, 10, 15])
        plot_traj(qqq[3], A, [7, 9, 14])
        plot_traj(qqq[4], A, [13, 6, 11])
    
    else:
        raise ValueError('At the moment only plots one- and two-qubit trajectories.')



def pcolor(W, a, b, clim=(0, 1)):
    """Easy pseudocolor plot.

    Plots the 2D function given in the matrix W.
    The vectors x and y define the coordinate grid.
    clim is an optional parameter for color limits.

    Returns the plot object.
    """
    # Ville Bergholm 2010

    # a and b are quad midpoint coordinates but pcolor wants quad vertices, so
    def to_quad(x):
        "Quad midpoints to vertices."
        return (r_[x, x[-1]] + r_[x[0], x]) / 2

    plt.gcf().clf()  # clear the figure
    p = plt.pcolor(to_quad(a), to_quad(b), W, clim = clim, cmap = asongoficeandfire())
    plt.axis('equal')
    plt.axis('tight')
    #shading('interp')
    plt.colorbar()
    return p


def makemovie(filename, frameset, plot_func, *arg):
    """Create an AVI movie. FIXME
    aviobj = makemovie(filename, frameset, plot_func [, ...])

    Creates an AVI movie file named 'filename.avi' in the current directory.
    Frame k in the movie is obtained from the contents of the
    current figure after calling plot_func(frameset[k]).
    The optional extra parameters are passed directly to avifile.

    Returns the closed avi object handle.

    Example: makemovie('test', cell_vector_of_states, @(x) plot(x))
    """
    # James D. Whitfield 2009
    # Ville Bergholm 2009-2010

    # create an AVI object
    aviobj = avifile(filename, arg)

    fig = figure('Visible', 'off')
    for k in frameset:
        plot_func(k)
        aviobj = addframe(aviobj, fig)
        #  F = getframe(fig)   
        #  aviobj = addframe(aviobj, F)

    close(fig)
    aviobj = close(aviobj)


@copy_memoize
def asongoficeandfire(n=127):
    """Colormap with blues and reds. Wraps.

    Returns a matplotlib.colors.Colormap object.
    n is the number of color definitions in the map.
    """
    # Ville Bergholm 2010-2011

    from matplotlib import colors
    # exponent
    d = 3.1
    p = linspace(-1, 1, n)
    # negative values: reds
    x = p[p < 0]
    c = c_[1 -((1+x) ** d), 0.5*(tanh(4*(-x -0.5)) + 1), (-x) ** d]
    # positive values: blues
    x = p[p >= 0]
    c = r_[c, c_[x ** d, 0.5*(tanh(4*(x -0.5)) + 1), 1 -((1-x) ** d)]]
    return colors.ListedColormap(c, name='asongoficeandfire')
    # TODO colors.LinearSegmentedColormap(name, segmentdata, N=256, gamma=1.0)



# -*- coding: utf-8 -*-
"""
Linear maps (:mod:`qit.lmap`)
=============================

Bounded finite-dimensional linear maps are represented using :class:`lmap` class instances.
In addition to the matrix representing the map, they contain
the dimension vectors of the domain and codomain vector spaces.
All the usual scalar-map and map-map arithmetic operators are
provided, including the exponentiation of maps by integers.


.. currentmodule:: qit.lmap.lmap


Utilities
---------

.. autosummary::

   remove_singletons
   is_compatible
   is_ket


Linear algebra
--------------

.. autosummary::

   conj
   transpose
   ctranspose
   trace
   norm
   reorder


Non-member functions:

.. currentmodule:: qit.lmap

.. autosummary::
   tensor
""" 
# Ville Bergholm 2008-2011

from __future__ import division, absolute_import, print_function, unicode_literals

from copy import copy, deepcopy

import numpy as np
import scipy.sparse as ssp

from .base import tol

__all__ = ['numstr_to_array', 'array_to_numstr', 'lmap', 'tensor']



def numstr_to_array(s):
    """Utility, converts a numeric string to the corresponding array."""
    return np.array([ord(x) -ord('0') for x in s])


def array_to_numstr(s):
    """Utility, converts an integer array to the corresponding numeric string."""
    return "".join([chr(x +ord('0')) for x in s])



class lmap(object):
    """Bounded linear maps between tensor products of finite-dimensional Hilbert spaces.

    Contains both the order-2 tensor and the dimensional information.

    TODO Another possible interpretation of lmap would be to
    treat each subsystem as an index, with the subsystems within dim{1} and dim{2}
    corresponding to contravariant and covariant indices, respectively?

    Variables:
    data:  ndarray of tensor data
    dim:   tuple of input and output dimension tuples, big-endian: ((out), (in))

    Base class of state.
    """

# TODO def __format__(self, format_spec)
# TODO linalg efficiency: copy vs. view

    def __init__(self, s, dim=None):
        """Construct an lmap.

        s:    ndarray OR valid initializer for ndarray OR lmap instance
              A copy is made unless s is an ndarray.

        dim:  2-tuple containing the output and input subsystem dimensions
              stored in tuples:  dim == ((out), (in)).
              If dim, (out) or (in) is None, the corresponding dimensions
              are inferred from s.
        
        calling syntax                     resulting dim
        ==============                     =============
        lmap(rand(a))                      ((a,), (1,))      1D array default: ket vector
        lmap(rand(a), ((1,), None))        ((1,), (a,))      bra vector given as a 1D array
        lmap(rand(a,b))                    ((a,), (b,))      2D array, all dims inferred
        lmap(rand(4,b), ((2, 2), None))    ((2, 2), (b,))    2D array, output: two qubits 
        lmap(rand(a,6), (None, (3, 2)))    ((a,), (3, 2))    2D array, input: qutrit+qubit
        lmap(rand(6,6), ((3, 2), (2, 3)))  ((3, 2), (2, 3))  2D array, all dims given

        lmap(A)             (A is an lmap) copy constructor
        lmap(A, dim)        (A is an lmap) copy constructor, redefine the dimensions
        """
        # initialize the ndarray part
        if isinstance(s, lmap):
            # copy constructor
            self.data = deepcopy(s.data)
            defdim = s.dim  # copy the dimensions too, unless redefined
        else:
            if ssp.isspmatrix(s):
                # TODO FIXME handle sparse matrices properly
                # TODO lmap constructor, mul/add, tensor funcs must be able to handle both dense and sparse arrays.
                s = s.todense()

            # valid array initializer
            self.data = np.asarray(s) # NOTE that if s is an ndarray it is not copied here

            # into a 2d array
            if self.data.ndim == 0:
                # scalar
                self.data.resize((1, 1))
            elif self.data.ndim == 1:
                # vector, ket by default
                self.data.resize((self.data.size, 1))
            elif self.data.ndim > 2:
                raise ValueError('Array dimension must be <= 2.')
            # now self.data.ndim == 2, always

            # is it a bra given as a 1D array?
            if dim and dim[0] == (1,):
                self.data.resize((1, self.data.size))

            # infer default dims from data (wrap them in tuples!)
            defdim = tuple([(k,) for k in self.data.shape])

        # set the dimensions
        if dim == None:
            # infer both dimensions from s
            dim = (None, None)

        self.dim = []
        for k in range(len(dim)):
            if dim[k] == None:
                # not specified, use default
                self.dim.append(defdim[k])
            else:
                self.dim.append(tuple(dim[k]))
        self.dim = tuple(self.dim)

        # check dimensions
        if self.data.shape != tuple(map(np.prod, self.dim)):
            raise ValueError('Dimensions of the array do not match the combined dimensions of the subsystems.')


    def __repr__(self):
        """Display the lmap in a neat format."""
        out = ''
        # is it a vector? (a map with a singleton domain or codomain dimension)
        sh = self.data.shape
        if 1 in sh:
            # vector 
            # ket or bra?
            if sh[1] == 1:
                # let scalars be kets too
                dim = self.dim[0]
                is_ket = True
            else:
                dim = self.dim[1]
                is_ket = False

            # loop over all vector elements
            printed = 0
            d = np.prod(dim)
            for ind in range(d):
                # TODO with sparse arrays we could do better
                # sanity check, do not display lmaps with hundreds of terms
                if ind >= 128 or printed >= 20:
                    out += ' ...'
                    break

                temp = self.data.flat[ind]
                # make sure there is something to print
                if abs(temp) < tol:
                    continue

                printed += 1
                if abs(temp.imag) < tol:
                    # just the real part
                    out += ' {0:+.4g}'.format(temp.real)
                elif abs(temp.real) < tol:
                    # just the imaginary part
                    out += ' {0:+.4g}j'.format(temp.imag)
                else:
                    # both
                    out += ' +({0:.4g}{1:+.4g}j)'.format(temp.real, temp.imag) #' +' + str(temp)

                # ket or bra symbol
                temp = array_to_numstr(np.unravel_index(ind, dim))
                if is_ket:
                    out += ' |' + temp + '>'
                else:
                    out += ' <' + temp + '|'
        else:
            # matrix
            out = self.data.__repr__()

        out += '\ndim: ' + str(self.dim[0]) + ' <- ' + str(self.dim[1])
        return out


# utilities

    def _inplacer(self, inplace):
        """Utility for implementing inplace operations.

        Functions using this should begin with s = self._inplacer(inplace)
        and end with return s
        """
        if inplace:
            return self
        else:
            return deepcopy(self)


    def remove_singletons(self):
        """Eliminate unnecessary singleton dimensions.

        NOTE: changes the object itself!
        """
        dd = []
        for d in self.dim:
            temp = tuple([x for x in d if x > 1])
            if len(temp) == 0:
                temp = (1,)
            dd.append(temp)
        self.dim = tuple(dd)
        return


    def is_compatible(self, t):
        """True iff the lmaps have equal dimensions and can thus be added."""
        if not isinstance(t, lmap):
            raise TypeError('t is not an lmap.')
        return self.dim == t.dim


    def is_ket(self):
        """True if the lmap is a ket."""
        return self.data.shape[1] == 1


# linear algebra

    def conj(self):
        """Complex conjugate."""
        s = copy(self)  # preserves the type, important for subclasses
        s.data = np.conj(self.data) # copy
        return s


    def transpose(self):
        """Transpose."""
        s = copy(self)
        s.dim = (s.dim[1], s.dim[0]) # swap dims
        s.data = self.data.transpose().copy()
        return s


    def ctranspose(self):
        """Hermitian conjugate."""
        s = copy(self)
        s.dim = (s.dim[1], s.dim[0]) # swap dims
        s.data = np.conj(self.data).transpose() # view to a copy
        return s


    def __mul__(self, t):
        """Multiplication of lmaps by lmaps and scalars."""
        # must be able to handle sparse data
        if isinstance(t, lmap):
            if self.dim[1] != t.dim[0]:
                raise ValueError('The dimensions do not match.')
            else:
                s = copy(self)
                s.dim = (self.dim[0], t.dim[1])
                s.data = self.data.dot(t.data)
        else:
            # t is a scalar
            s = copy(self)
            s.data = self.data * t
        return s


    def __rmul__(self, t):
        """Multiplication of lmaps by scalars, reverse."""
        # scalars commute, lmaps already handled by __mul__
        return self.__mul__(t)


    def __div__(self, t):
        """Division of lmaps by scalars from the right."""
        s = copy(self)
        s.data = self.data / t
        return s


    def __truediv__(self, t):
        """Division of lmaps by scalars from the right."""
        s = copy(self)
        s.data = self.data / t
        return s


    def __add__(self, t):
        """Addition of lmaps."""
        if not self.is_compatible(t):
            raise ValueError('The lmaps are not compatible.')
        s = copy(self)
        s.data = self.data + t.data
        return s


    def __sub__(self, t):
        """Subtraction of lmaps."""
        if not self.is_compatible(t):
            raise ValueError('The lmaps are not compatible.')
        s = copy(self)
        s.data = self.data - t.data
        return s


    def __pow__(self, n):
        """Exponentiation of lmaps by integer scalars."""
        if self.dim[0] != self.dim[1]:
            raise ValueError('The dimensions do not match.')
        s = copy(self)
        s.data = np.linalg.matrix_power(self.data, n)
        return s


    def __imul__(self, t):
        """In-place multiplication of lmaps by scalars from the right."""
        self.data *= t
        return self


    def __itruediv__(self, t):
        """In-place division of lmaps by scalars from the right."""
        self.data /= t
        return self


    def __iadd__(self, t):
        """In-place addition of lmaps."""
        if not self.is_compatible(t):
            raise ValueError('The lmaps are not compatible.')
        self.data += t.data
        return self


    def __isub__(self, t):
        """In-place subtraction of lmaps."""
        if not self.is_compatible(t):
            raise ValueError('The lmaps are not compatible.')
        self.data -= t.data
        return self


    def trace(self):
        """Trace of the lmap.

        The trace is only properly defined if self.dim[0] == self.dim[1].
        """
        if not np.array_equal(self.dim[0], self.dim[1]):
            raise ValueError('Trace not defined for non-endomorphisms.')

        return np.trace(self.data)


    def norm(self):
        """Matrix norm of the lmap."""
        return np.linalg.norm(self.data)


# subsystem ordering

    def reorder(self, perm, inplace=False):
        """Change the relative order of the input and/or output subsystems.

        Returns a copy of the lmap with permuted subsystem order.

        A permutation can be either None (do nothing), a pair (a, b) of subsystems to be swapped,
        or a tuple containing a full permutation of the subsystems.
        Two subsystems to be swapped must be in decreasing order so as not
        to mistake the full identity permutation (0, 1) for a swap.

        reorder((None, (2, 1, 0)))   ignore first index, reverse the order of subsystems in the second
        reorder(((5, 2), None))      swap the subsystems 2 and 5 in the first index, ignore the second

        NOTE: The full permutations are interpreted in the same sense as
        numpy.transpose() understands them, i.e. the permutation
        tuple is the new ordering of the old subsystem indices.
        This is the inverse of the mathematically more common "one-line" notation.
        """
        s = self._inplacer(inplace)

        orig_d = s.data.shape  # original dimensions
        total_d = []
        total_perm = []
        last_used_index = 0
        newdim = list(s.dim)

        # loop over indices
        for k, this_perm in enumerate(perm):
            # avoid a subtle problem with the input syntax, (0, 1) must not be understood as swap!
            if this_perm != None and tuple(this_perm) == (0, 1):
                this_perm = None

            # requested permutation for this index
            if this_perm == None:
                # no change
                # let the dimensions vector be, lump all subsystems in this index into one
                this_dim = (orig_d[k],)
                this_perm = np.array([0])
                this_n = 1
            else:
                this_dim  = np.array(s.dim[k])  # subsystem dims
                this_perm = np.array(this_perm) # requested permutation for this index
                this_n = len(this_dim)  # number of subsystems

                temp = np.arange(this_n) # identity permutation

                if len(this_perm) == 2:
                    # swap two subsystems
                    temp[this_perm] = this_perm[::-1]
                    this_perm = temp
                else:
                    # full permutation
                    if len(set(temp) ^ set(this_perm)) != 0:
                        raise ValueError('Invalid permutation.')

                # reorder the dimensions vector
                newdim[k] = tuple(this_dim[this_perm])

            # big-endian ordering
            total_d.extend(this_dim)
            total_perm.extend(last_used_index + this_perm)
            last_used_index += this_n

        # tensor into another tensor which has one index per subsystem, permute dimensions, back into a tensor with the original number of indices
        s.dim = tuple(newdim)
        s.data = s.data.reshape(total_d).transpose(total_perm).reshape(orig_d)
        return s



def tensor(*arg):
    """Tensor product of lmaps."""
    data = 1
    dout = []
    din  = []

    for k in arg:
        # concatenate dimensions
        dout += k.dim[0]
        din  += k.dim[1]
        # kronecker product of the data
        data = np.kron(data, k.data)

    s = lmap(data, (tuple(dout), tuple(din)))
    return s



# -*- coding: utf-8 -*-
"Unit tests for qit.utils"
# Ville Bergholm 2009-2014

import unittest
from numpy import mat, array, dot, eye, trace
from numpy.random import rand, randn
from numpy.linalg import norm, det, eigvalsh
from scipy.linalg import expm

# HACK to import the module in this source distribution, not the installed one!
import sys, os
sys.path.insert(0, os.path.abspath('.'))

from qit import version
from qit.base  import tol
from qit.utils import *


def randn_complex(*arg):
    "Returns an array of random complex numbers, normally distributed."
    return randn(*arg) +1j*randn(*arg)



class UtilsTest(unittest.TestCase):
    def setUp(self):
        pass


    def assertHermitian(self, H, delta):
        "Make sure H is hermitian."
        dim = len(H)
        H = mat(H)
        self.assertAlmostEqual(norm(H -H.H), 0, delta=delta)


    def assertUnitary(self, U, delta):
        "Make sure U is unitary."
        dim = len(U)
        U = mat(U)
        self.assertAlmostEqual(norm(U * U.H -eye(dim)), 0, delta=delta)
        self.assertAlmostEqual(norm(U.H * U -eye(dim)), 0, delta=delta)
        

    def test_funcs(self):
        "Testing the utils module."

        dim = 10

        ### expv
        A = randn_complex(dim, dim)
        H = rand_hermitian(dim)
        v = randn_complex(dim)
        newtol = 1e2 * tol  # use a larger tolerance here

        # arbitrary matrix with Arnoldi iteration
        w, err, hump = expv(1, A, v, m = dim // 2)
        self.assertAlmostEqual(norm(w - dot(expm(1*A), v)), 0, delta=newtol)
        w, err, hump = expv(1, A, v, m = dim)  # force a happy breakdown
        self.assertAlmostEqual(norm(w - dot(expm(1*A), v)), 0, delta=newtol)

        # Hermitian matrix with Lanczos iteration
        w, err, hump = expv(1, H, v, m = dim // 2, iteration='lanczos')
        self.assertAlmostEqual(norm(w - dot(expm(1*H), v)), 0, delta=newtol)
        w, err, hump = expv(1, H, v, m = dim, iteration = 'lanczos')
        self.assertAlmostEqual(norm(w - dot(expm(1*H), v)), 0, delta=newtol)

        # FIXME why does Lanczos work with nonhermitian matrices?
        w, err, hump = expv(1, A, v, m = dim // 2, iteration='lanczos')
        self.assertAlmostEqual(norm(w - dot(expm(1*A), v)), 0, delta=newtol)
        w, err, hump = expv(1, A, v, m = dim, iteration = 'lanczos')
        self.assertAlmostEqual(norm(w - dot(expm(1*A), v)), 0, delta=newtol)


        dim = 5

        ### random matrices
        H = rand_hermitian(dim)
        self.assertHermitian(H, delta=tol)

        U = rand_U(dim)
        self.assertUnitary(U, delta=tol)

        U = rand_SU(dim)
        self.assertUnitary(U, delta=tol)
        self.assertAlmostEqual(det(U), 1, delta=tol) # det 1

        rho = rand_positive(dim)
        self.assertHermitian(rho, delta=tol)
        self.assertAlmostEqual(trace(rho), 1, delta=tol) # trace 1
        temp = eigvalsh(rho)
        self.assertAlmostEqual(norm(temp.imag), 0, delta=tol) # real eigenvalues
        self.assertAlmostEqual(norm(temp - abs(temp)), 0, delta=tol) # nonnegative eigenvalues


        ### superoperators
        L = mat(rand_U(dim))
        R = mat(rand_U(dim))
        v = vec(array(rho))
        self.assertAlmostEqual(norm(rho -inv_vec(v)), 0, delta=tol)
        self.assertAlmostEqual(norm(L*rho*R -inv_vec(dot(lrmul(L, R), v))), 0, delta=tol)
        self.assertAlmostEqual(norm(L*rho -inv_vec(dot(lmul(L), v))), 0, delta=tol)
        self.assertAlmostEqual(norm(rho*R -inv_vec(dot(rmul(R), v))), 0, delta=tol)


        ### physical operators
        J = angular_momentum(dim)
        self.assertEqual(len(J), 3)  #  3 components
        for A in J:
            self.assertHermitian(A, delta=tol)
        self.assertAlmostEqual(norm(comm(J[0], J[1]) - 1j*J[2]), 0, delta=tol)  # [Jx, Jy] == i Jz

        a = mat(boson_ladder(dim))
        temp = comm(a, a.H)
        self.assertAlmostEqual(norm(temp[:-1, :-1] -eye(dim-1)), 0, delta=tol)  # [a, a'] == I  (truncated, so skip the last row/col!)

        fff = fermion_ladder(3)
        # {f_j, f_k} = 0
        # {f_j, f_k^\dagger} = I \delta_{jk}
        for f in fff:
            fp = f.conj().transpose()
            self.assertAlmostEqual(norm(acomm(f, f)), 0, delta=tol)
            self.assertAlmostEqual(norm(acomm(f, fp) -eye(8)), 0, delta=tol)
            for j in fff:
                if not f is j:
                    self.assertAlmostEqual(norm(acomm(j, f)), 0, delta=tol)
                    self.assertAlmostEqual(norm(acomm(j, fp)), 0, delta=tol)

        ### SU(2) rotations


        ### spectral decomposition
        E, P = spectral_decomposition(H)
        temp = 0
        for k in range(len(E)):
            temp += E[k] * P[k]
        self.assertAlmostEqual(norm(temp -H), 0, delta=tol)


        # tensor bases


        # majorization

        # op_list

        # plots



if __name__ == '__main__':
    print('Testing QIT version ' + version())
    unittest.main()



# -*- coding: utf-8 -*-
"""
Born-Markov noise (:mod:`qit.markov`)
=====================================


This module simulates the effects of a heat bath coupled to a quantum
system, using the Born-Markov approximation.

The treatment in this module mostly follows Ref. :cite:`BP`.



Contents
--------

.. currentmodule:: qit.markov

.. autosummary::

   ops
   lindblad_ops
   superop   
   bath


:class:`bath` methods
---------------------

.. currentmodule:: qit.markov.bath

.. autosummary::

   build_LUT
   S_func
   set_cutoff
   corr
   fit
"""
# Ville Bergholm 2011-2014

from __future__ import division, absolute_import, print_function, unicode_literals

from numpy import (array, sqrt, exp, sin, cos, arctan2, tanh, dot, argsort, pi,
    r_, linspace, logspace, searchsorted, inf, newaxis, unravel_index)
from scipy.linalg import norm
from scipy.integrate import quad
import scipy.constants as const

from .base import sx, sz, tol
from .utils import lmul, rmul, lrmul, rand_hermitian, superop_lindblad, spectral_decomposition


__all__ = ['bath', 'ops', 'lindblad_ops', 'superop']


class bath(object):
    r"""Markovian heat bath.

    Currently only one type of bath is supported, a bosonic
    canonical ensemble at absolute temperature T, with a
    single-term coupling to the system.
    The bath spectral density is Ohmic with a cutoff:

    .. math::

       J(\omega) = \hbar^2 \omega \mathrm{cut}(\omega) \Theta(\omega)

    Two types of cutoffs are supported: exponential and sharp.

    .. math::

       \mathrm{cut}_{\text{exp}}(\omega)   &= \exp(-\omega / \omega_c),\\
       \mathrm{cut}_{\text{sharp}}(\omega) &= \Theta(\omega_c - \omega)

    The effects of the bath on the system are contained in the complex spectral correlation tensor

    .. math::

       \Gamma(\omega) = \frac{1}{2} \gamma(\omega) +i S(\omega)

    where :math:`\gamma` and :math:`S` are real.
    Computing values of this tensor is the main purpose of this class.

    .. math::

       \gamma(\omega) &= \frac{2 \pi}{\hbar^2} (J(\omega) -J(-\omega))(1 + n(\omega)),\\
       S(\omega) &= \frac{1}{\hbar^2} \int_0^\infty \mathrm{d}\nu J(\nu) \frac{\omega \coth(\beta \hbar \nu/2) +\nu}{\omega^2 -\nu^2}.


    where :math:`n(\omega) := 1/(e^{\beta \hbar \omega} - 1)` is the Planc function and :math:`\beta = 1/(k_B T)`. 
    Since :math:`\Gamma` is pretty expensive to compute, we store the computed results into a lookup table which is used to interpolate nearby values.


    Public data:

    ===========  ===========
    Data member  Description
    ===========  ===========
    type         Bath type. Currently only 'ohmic' is supported.
    omega0       Energy scale (in Hz). :math:`\hbar \omega_0` is the unit of energy for all Hamiltonians related to this bath.
    T            Absolute temperature of the bath (in K).
    scale        Dimensionless temperature scaling parameter :math:`\hbar \omega_0 / (k_B T)`.
    cut_type     Spectral density cutoff type (string).
    cut_limit    Spectral density cutoff limit :math:`\omega_c / \omega_0`.
    ===========  ===========    


    Private data (set automatically):

    ===========  ===========
    Data member  Description
    ===========  ===========
    cut_func     Spectral density cutoff function.
    j            Spectral density profile. :math:`J(\omega_0 x)/\omega_0 = \hbar^2 j(x) \mathrm{cut\_func}(x) \Theta(x)`.
    g_func       Spectral correlation tensor, real part. :math:`\gamma(\omega_0 x) / \omega_0 = \mathrm{g\_func}(x) \mathrm{cut\_func}(x)`. For the imaginary part, see :func:`S_func`.
    g0           :math:`\lim_{\omega \to 0} \gamma(\omega)`.
    s0           :math:`\lim_{\omega \to 0} S(\omega)`.
    dH           Lookup table.
    gs_table     Lookup table. :math:`(\gamma / S)(\omega_0 \text{dH[k]}) / \omega_0` = gs_table[k, (0/1)].
    ===========  ===========
    """
    # Ville Bergholm 2009-2011

    def __init__(self, type, omega0, T):
        """constructor
    
        Sets up a descriptor for a heat bath coupled to a quantum system.
        """
        # basic bath parameters
        self.type   = type
        self.omega0 = omega0
        self.T      = T
        # shorthand
        self.scale = const.hbar * omega0 / (const.k * T)

        if type == 'ohmic':
            # Ohmic bosonic bath, canonical ensemble, single-term coupling
            self.g_func = lambda x: 2 * pi * x * (1 + 1 / (exp(self.scale * x) - 1))
            self.g0 = 2 * pi / self.scale  # limit of g at x == 0
            self.j = lambda x: x
        else:
            raise ValueError('Unknown bath type.')

        # defaults, can be changed later
        self.set_cutoff('sharp', 20)


    def __repr__(self):
        """String representation."""
        return """Markovian heat bath.  Spectral density: {sd}, T = {temp:g}, omega0 = {omega:g}""".format(sd = self.type, temp = self.T, omega = self.omega0)


    def build_LUT(self):
        """Build a lookup table for the S integral. Unused.
        """
        raise RuntimeError('unused')
        # TODO justify limits for S lookup
        if limit > 10:
            temp = logspace(log10(10.2), log10(limit), 50)
            temp = r_[linspace(0.1, 10, 100), temp] # sampling is denser near zero, where S changes more rapidly
        else:
            temp = linspace(0.1, limit, 10)
        self.dH = r_[-temp[::-1], 0, temp]

        self.s_table = []
        for k in range(len(self.dH)):
            self.s_table[k] = self.S_func(self.dH[k])

        # limits at inifinity
        self.dH      = r_[-inf, self.dH, inf]
        self.s_table = r_[0, self.s_table, 0]

        plot(self.dH, self.s_table, 'k-x')


    def S_func(self, x):
        r"""Spectral correlation tensor, imaginary part.

        .. math::

           \mathrm{S\_func}(x) = S(x \omega_0) / \omega_0
           = \frac{1}{\hbar^2 \omega_0} P\int_0^\infty \mathrm{d}\nu J(\omega_0 \nu) \frac{x \coth(\nu \mathrm{scale}/2) +\nu}{x^2 -\nu^2}.
        """
        ep = 1e-5 # epsilon for Cauchy principal value
        if abs(x) <= 1e-8:
            return self.s0
        else:
            # Cauchy principal value, integrand has simple poles at \nu = \pm x.
            f = lambda nu: self.j(nu) * self.cut_func(nu) * (x / tanh(self.scale * nu / 2) + nu) / (x**2 -nu**2)
            a, abserr = quad(f, ep, abs(x) - ep) 
            b, abserr = quad(f, abs(x) + ep, inf) # 100 * self.cut_limit)
            return a + b


    def set_cutoff(self, type, lim):
        """Set the spectral density cutoff."""
        self.cut_type = type
        self.cut_limit = lim  # == omega_c/omega0

        # update cutoff function
        if self.cut_type == 'sharp':
            self.cut_func = lambda x: abs(x) <= self.cut_limit  # Heaviside theta cutoff
        elif self.cut_type == 'exp':
            self.cut_func = lambda x: exp(-abs(x) / self.cut_limit)  # exponential cutoff
        else:
            raise ValueError('Unknown cutoff type "{0}"'.format(self.cut_type))

        if self.type == 'ohmic':
            self.s0 = -self.cut_limit  # limit of S at dH == 0
  
        # clear lookup tables, since changing the cutoff requires recalc of S
        # start with a single point precomputed
        self.dH = array([-inf, 0, inf])
        self.gs_table = array([[0, 0], [self.g0, self.s0], [0, 0]])



    def corr(self, x):
        r"""Bath spectral correlation tensor.
        ::

          g, s = corr(x)

        Returns the bath spectral correlation tensor :math:`\Gamma` evaluated at :math:`\omega_0 x`:

        .. math::

           \Gamma(\omega_0 x) / \omega_0 = \frac{1}{2} g +i s
        """
        # Ville Bergholm 2009-2011

        tol = 1e-8
        max_w = 0.1 # maximum interpolation distance, TODO justify

        # assume parameters are set and lookup table computed
        #s = interp1(self.dH, self.s_table, x, 'linear', 0)

        # TODO dH and gs_table into a single dictionary?
        # binary search for the interval [dH_a, dH_b) in which x falls
        b = searchsorted(self.dH, x, side = 'right')
        a = b - 1
        ee = self.dH[[a, b]]
        tt = self.gs_table[[a, b], :]
        # now x is in [ee[0], ee[1])

        gap = ee[1] - ee[0]
        d1 = abs(x - ee[0])
        d2 = abs(x - ee[1])

        def interpolate(ee, tt, x):
            "Quick interpolation."
            # interp1 does way too many checks
            return tt[0] + ((x - ee[0]) / (ee[1] - ee[0])) * (tt[1] - tt[0])

        # x close enough to either endpoint?
        if d1 <= tol:
            return self.gs_table[a, :]
        elif d2 <= tol:
            return self.gs_table[b, :]
        elif gap <= max_w + tol:  # short enough gap to interpolate?
            return interpolate(ee, tt, x)
        else: # compute a new point p, then interpolate
            if gap <= 2 * max_w:
                p = ee[0] + gap / 2 # gap midpoint
                if x < p:
                    idx = 1 # which ee p will replace
                else:
                    idx = 0
            elif d1 <= max_w: # x within interpolation distance from one of the gap endpoints?
                p = ee[0] + max_w
                idx = 1
            elif d2 <= max_w:
                p = ee[1] - max_w
                idx = 0
            else: # x not near anything, don't interpolate
                p = x
                idx = 0

            # compute new g, s values at p and insert them into the table
            s = self.S_func(p)
            if abs(p) <= tol:
                g = self.g0 # limit at p == 0
            else:
                g = self.g_func(p) * self.cut_func(p)
            temp = array([[g, s]])

            self.dH = r_[self.dH[:b], p, self.dH[b:]]
            self.gs_table = r_[self.gs_table[:b], temp, self.gs_table[b:]]

            # now interpolate the required value
            ee[idx] = p
            tt[idx, :] = temp
            return interpolate(ee, tt, x)



    def fit(self, delta, T1, T2):
        r"""Qubit-bath coupling that reproduces given decoherence times.
        ::

          H, D = fit(delta, T1, T2)

        Returns the qubit Hamiltonian H and the qubit-bath coupling operator D
        that reproduce the decoherence times T1 and T2 (in units of :math:`1/\omega_0`)
        for a single-qubit system coupled to the bath.
        delta is the energy splitting for the qubit (in units of :math:`\hbar \omega_0`).

        The bath object is not modified.
        """
        # Ville Bergholm 2009-2010

        if self.type == 'ohmic':
            # Fitting an ohmic bath to a given set of decoherence times

            iTd = 1 / T2 -0.5 / T1 # inverse pure dephasing time
            if iTd < 0:
                raise ValueError('Unphysical decoherence times!')
    
            # match bath couplings to T1, T2
            temp = self.scale * delta / 2
            alpha = arctan2(1, sqrt(T1 * iTd / tanh(temp) * temp * self.cut_func(delta)))
            # dimensionless system-bath coupling factor squared
            N = iTd * self.scale / (4 * pi * cos(alpha)**2)

            # qubit Hamiltonian
            H = -delta/2 * sz

            # noise coupling
            D = sqrt(N) * (cos(alpha) * sz + sin(alpha) * sx)

            # decoherence times in scaled time units
            #T1 = 1/(N * sin(alpha)**2 * 2*pi * delta * coth(temp) * self.cut_func(delta))
            #T_dephase = self.scale/(N *4*pi*cos(alpha)**2)
            #T2 = 1/(0.5/T1 +1/T_dephase)
        else:
            raise NotImplementedError('Unknown bath type.')
        return H, D



def ops(H, D):
    r"""Jump operators for a Born-Markov master equation.
    ::

      dH, A = ops(H, D)

    Builds the jump operators for a Hamiltonian operator H and
    a (Hermitian) interaction operator D.

    Returns (dH, A), where dH is a list of the sorted unique nonnegative differences between
    eigenvalues of H, and A is a sequence of the corresponding jump operators:
    :math:`A_k(dH_i) = A[k][i]`.

    Since :math:`A_k(-dH) = A_k^\dagger(dH)`, only the nonnegative dH:s and corresponding A:s are returned.
    """
    # Ville Bergholm 2009-2011

    E, P = spectral_decomposition(H)
    m = len(E) # unique eigenvalues
    # energy difference matrix is antisymmetric, so we really only need the lower triangle
    deltaE = E[:, newaxis] - E  # deltaE[i,j] == E[i] - E[j]

    # mergesort is a stable sorting algorithm
    ind = argsort(deltaE, axis = None, kind = 'mergesort')
    # index of first lower triangle element
    s = m * (m - 1) / 2
    #assert(ind[s], 0)
    ind = ind[s:] # lower triangle indices only
    deltaE = deltaE.flat[ind] # lower triangle flattened

    if not isinstance(D, (list, tuple)):
        D = [D] # D needs to be a sequence, even if it has just one element
    n_D = len(D) # number of bath coupling ops

    # combine degenerate deltaE, build jump ops
    A = []
    # first dH == 0
    dH = [deltaE[0]]
    r, c = unravel_index(ind[0], (m, m))
    for d in D:
        A.append( [ dot(dot(P[c], d), P[r]) ] )

    for k in range(1, len(deltaE)):
        r, c = unravel_index(ind[k], (m, m))
        if abs(deltaE[k] - deltaE[k-1]) > tol:
            # new omega value, new jump op
            dH.append(deltaE[k])
            for op in range(n_D):
                A[op].append( dot(dot(P[c], D[op]), P[r]) )
        else:
            # extend current op
            for op in range(n_D):
                A[op][-1] += dot(dot(P[c], D[op]), P[r])

    return dH, A



def _check_baths(B):
    """Internal helper."""
    if not isinstance(B, (list, tuple)):
        B = [B] # needs to be a list, even if it has just one element

    # make sure the baths have the same omega0!
    temp = B[0].omega0
    for k in B:
        if k.omega0 != temp:
            raise ValueError('All the baths must have the same energy scale omega0!')
    return B



def lindblad_ops(H, D, B):
    r"""Lindblad operators for a Born-Markov master equation.
    ::

       L, H_LS = lindblad_ops(H, D, B)

    Builds the Lindblad operators corresponding to a
    base Hamiltonian H and a (Hermitian) interaction operator D
    coupling the system to bath B.

    Returns :math:`L = \{A_i / \omega_0 \}_i` and :math:`H_{\text{LS}} / (\hbar \omega_0)`,
    where :math:`A_i` are the Lindblad operators and :math:`H_{\text{LS}}` is the Lamb shift.

    B can also be a list of baths, in which case D has to be
    a list of the corresponding interaction operators.
    """
    # Ville Bergholm 2009-2011

    B = _check_baths(B)

    # jump ops
    dH, X = ops(H, D)
    H_LS = 0
    L = []
    for n, b in enumerate(B):
        A = X[n] # jump ops for bath/interaction op n

        # dH == 0 terms
        g, s = b.corr(0)
        L.append(sqrt(g) * A[0])
        H_LS += s * dot(A[0].conj().transpose(), A[0])  # Lamb shift

        for k in range(1, len(dH)):
            # first the positive energy shift
            g, s = b.corr(dH[k])
            L.append(sqrt(g) * A[k])
            H_LS += s * dot(A[k].conj().transpose(), A[k])

            # now the corresponding negative energy shift
            g, s = b.corr(-dH[k])
            L.append(sqrt(g) * A[k].conj().transpose())   # note the difference here, A(-omega) = A'(omega)
            H_LS += s * dot(A[k], A[k].conj().transpose()) # here too

    return L, H_LS
    # TODO ops for different baths can be combined into a single basis,
    # N^2-1 ops max in total



def superop(H, D, B):
    r"""Liouvillian superoperator for a Born-Markov master equation.

    Builds the Liouvillian superoperator L corresponding to a
    base Hamiltonian H and a (Hermitian) interaction operator D
    coupling the system to bath B.

    Returns :math:`L/\omega_0`, which includes the system Hamiltonian, the Lamb shift,
    and the Lindblad dissipator.

    B can also be a list of baths, in which case D has to be
    a list of the corresponding interaction operators.
    """
    # Ville Bergholm 2009-2011

    B = _check_baths(B)

    # jump ops
    dH, X = ops(H, D)
    iH_LS = 1j * H  # i * (system Hamiltonian + Lamb-Stark shift)
    acomm = 0
    diss = 0
    for n, b in enumerate(B):
        A = X[n] # jump ops for bath/interaction op n

        # we build the Liouvillian in a funny order to be a bit more efficient
        # dH == 0 terms
        [g, s] = b.corr(0)
        temp = dot(A[0].conj().transpose(), A[0])

        iH_LS += (1j * s) * temp  # Lamb shift
        acomm += (-0.5 * g) * temp # anticommutator
        diss  += lrmul(g * A[0], A[0].conj().transpose()) # dissipator (part)

        for k in range(1, len(dH)):
            # first the positive energy shift
            g, s = b.corr(dH[k])
            temp = dot(A[k].conj().transpose(), A[k])
            iH_LS += (1j * s) * temp
            acomm += (-0.5 * g) * temp
            diss  += lrmul(g * A[k], A[k].conj().transpose())

            # now the corresponding negative energy shift
            g, s = b.corr(-dH[k])
            temp = dot(A[k], A[k].conj().transpose()) # note the difference here, A(-omega) = A'(omega)
            iH_LS += (1j * s) * temp
            acomm += (-0.5 * g) * temp
            diss  += lrmul(g * A[k].conj().transpose(), A[k]) # here too

    return lmul(acomm -iH_LS) +rmul(acomm +iH_LS) +diss



# -*- coding: utf-8 -*-
"""Python Quantum Information Toolkit

See the README.txt file included in this distribution
or the project website at http://qit.sourceforge.net/
"""
# Ville Bergholm 2011-2014

from __future__ import division, absolute_import, print_function, unicode_literals


# toolkit version number
__version__ = '0.11.0'

def version():
    """Returns the QIT version number (as a string)."""
    return __version__


import numpy as np
import scipy as sp
import scipy.constants as const
import matplotlib.pyplot as plt

from .base import *
from .lmap import *
from .utils import *
from .state import *
from .plot import *
from . import gate, hamiltonian, ho, invariant, markov, seq, examples

#print('Python Quantum Information Toolkit, version {0}.'.format(__version__))



# -*- coding: utf-8 -*-
"""
Quantum states (:mod:`qit.state`)
=================================


In QIT, quantum states are represented by the :class:`state` class,
defined in this module.


.. currentmodule:: qit.state.state


Utilities
---------

.. autosummary::

   check
   subsystems
   dims
   clean_selection
   invert_selection
   fix_phase
   normalize
   to_ket
   to_op
   trace
   ptrace
   ptranspose
   reorder
   tensor
   plot


Physics
-------

.. autosummary::

   ev
   var
   prob
   projector
   u_propagate
   propagate
   kraus_propagate
   measure
   

Quantum information
-------------------

.. autosummary::

   fidelity
   trace_dist
   purity
   schmidt
   entropy
   concurrence
   negativity
   lognegativity
   scott
   locc_convertible


Other state representations
---------------------------

.. autosummary::

   bloch_vector
   bloch_state
"""
# Ville Bergholm 2008-2012

from __future__ import division, absolute_import, print_function, unicode_literals

import collections
import numbers
import types
from copy import deepcopy

import numpy as np
from numpy import (array, asarray, empty, diag, sort, prod, cumsum, cumprod, exp, sqrt, trace,
    dot, vdot, roll, zeros, ones, r_, kron, isscalar, nonzero, ix_, linspace, meshgrid)
from numpy.random import rand, randn
import scipy as sp
from scipy.linalg import norm, eigh, eigvalsh, sqrtm, svd, svdvals, det
from scipy.integrate import ode

from .base import sy, Q_Bell, tol
from .lmap import tensor as lmap_tensor
from .lmap import numstr_to_array, array_to_numstr, lmap
from .utils import (_warn, vec, inv_vec, qubits, expv, rand_U, rand_SU, rand_positive, mkron,
    eighsort, spectral_decomposition, majorize, tensorbasis, assert_o)


__all__ = ['equal_dims', 'index_muls', 'state', 'fidelity', 'trace_dist']



def equal_dims(s, t):
    """True if s and t have equal dimensions."""
    return s.dims() == t.dims()


def index_muls(dim):
    """Index multipliers for C-ordered data"""
    # ravel_multi_index(s, dim) == dot(index_muls(dim), s)
    if len(dim) == 0:
        return array(())
    muls = roll(cumprod(dim[::-1]), 1)[::-1]
    muls[-1] = 1  # muls == [d_{n-1}*...*d_1, d_{n-1}*...*d_2, ..., d_{n-1}, 1]
    return muls



class state(lmap):
    """Class for quantum states.

    Describes the state (pure or mixed) of a discrete, possibly composite quantum system.
    The subsystem dimensions can be obtained with :func:`dims` (big-endian ordering).

    State class instances are special cases of lmaps. They have exactly two indices.
    If self.dim[1] == (1,), it is a ket representing a pure state.
    Otherwise both indices must have equal dimensions and the object represents a state operator.

    Does not require the state to be physical (it does not have to be trace-1, Hermitian, or nonnegative).
    """
    # Ville Bergholm 2008-2011
    # by default, all state methods leave self unchanged 

    def __init__(self, s, dim=None):
        """Construct a state.

        calling syntax            result
        ==============            ======
        state('00101')            standard basis ket |00101> in a five-qubit system
        state('02', (2, 3))       standard basis ket |02> in a qubit+qutrit system
        state(k, (2, 3))          linearized standard basis ket |k> in a qubit+qutrit system, k must be an integer scalar
        state(rand(4))            ket, infer dim[0] == (4,)
        state(rand(4), (2, 2))    ket, two qubits
        state(rand(4,4))          state operator, infer dim[0] == (4,)
        state(rand(6,6), (3, 2))  state operator, qutrit+qubit
        state('GHZ', (2, 2, 2))   named states (in this case the three-qubit GHZ state)

        state(s)                  (s is a state) copy constructor
        state(s, dim)             (s is a state) copy constructor, redefine the dimensions

        The currently supported named states are
          GHZ (Greenberger-Horne-Zeilinger),
          W,
          Bell1, Bell2, Bell3, Bell4 
        """
        from six import string_types

        # we want a tuple for dim
        if isinstance(dim, collections.Iterable):
            dim = tuple(dim)
        elif isscalar(dim):
            dim = (dim,)

        if isinstance(s, lmap):
            # copy constructor
            # state vector or operator? (also works with dim == None)
            if s.is_ket():
                dim = (dim, (1,))
            else:
                if s.dim[0] != s.dim[1]:
                    raise ValueError('State operator must be square.')
                dim = (dim, dim)
            # call the lmap copy constructor
            super(state, self).__init__(s, dim)
            return

        elif isinstance(s, string_types):
            # string

            if s[0].isalpha():
                # named state
                name = s.lower()

                if dim == None:
                    dim = (2, 2, 2) # default: three-qubit state

                n = len(dim) # subsystems
                s = zeros(prod(dim)) # ket
                dmin = min(dim)

                if name in ('bell1', 'bell2', 'bell3', 'bell4'):
                    # Bell state
                    dim = (2, 2)
                    s = deepcopy(Q_Bell[:, ord(name[-1]) - ord('1')])
                elif name == 'ghz':
                    # Greenberger-Horne-Zeilinger state
                    for k in range(dmin):
                        s[np.ravel_multi_index(n*(k,), dim)] = 1
                elif name == 'w':
                    # W state
                    ind = 1
                    for k in reversed(range(n)):
                        s[ind * (dim[k] - 1)] = 1
                        ind *= dim[k]
                else:
                    raise ValueError("Unknown named state '{0}'.".format(name))

                s /= norm(s) # normalize
      
            else:
                # number string defining a standard basis ket
                if dim == None:
                    n = len(s)  # number of subsystems
                    dim = qubits(n)  # assume they're qubits

                # calculate the linear index
                s = numstr_to_array(s)
                if any(s >= dim):
                    raise ValueError('Invalid basis ket.')

                ind = np.ravel_multi_index(s, dim)
                s = zeros(prod(dim)) # ket
                s[ind] = 1

            dim = (dim, (1,))  # ket

        elif isinstance(s, (numbers.Number, np.number)):
            # FIXME when numpy number hierarchy inherits the Python one
            # integer defining a standard basis ket
            if dim == None:
                raise ValueError('Need system dimension.')

            ind = s
            temp = prod(dim)  # total number of states
            if ind >= temp:
                raise ValueError('Invalid basis ket.')

            s = zeros(prod(dim)) # ket
            s[ind] = 1
            dim = (dim, (1,))  # ket

        else:
            # valid ndarray initializer representing a state vector or a state op
            s = array(s)
            if not 1 <= s.ndim <= 2:
                raise ValueError('State must be given as a state vector or a state operator.')

            # state vector or operator?
            if s.ndim == 2 and s.shape[1] != 1:
                if s.shape[0] != s.shape[1]:
                    raise ValueError('State operator matrix must be square.')
                dim = (dim, dim)  # op
            else:
                dim = (dim, (1,))  # ket

        # now s is an ndarray
        # call the lmap constructor
        super(state, self).__init__(s, dim)


# utility methods
# TODO design issue: for valid states, lots of these funcs should return reals (marked with a commented-out .real). should we just drop the imaginary part? what about if the state is invalid, how will the user know? what about numerical errors?
# TODO same thing except with normalization, should we assume states are normalized?

    def check(self):
        """Checks the validity of the state.

        Makes sure it is normalized, and if an operator, Hermitian and semipositive.
        """
        ok = True
        if abs(self.trace() - 1) > tol:
            _warn('State not properly normalized.')
            ok = False

        if not self.is_ket():
            if norm(self.data - self.data.conj().transpose()) > tol:
                _warn('State operator not Hermitian.')
                ok = False
            if min(eigvalsh(self.data)) < -tol:
                _warn('State operator not semipositive.')
                ok = False

        if not ok:
            raise ValueError('Not a valid state.')


    def subsystems(self):
        """Number of subsystems in the state."""
        return len(self.dim[0])


    def dims(self):
        """Dimensions of the state."""
        return self.dim[0] # dims of the other index must be equal (or 1)


    def clean_selection(self, sys):
        """Make a subsystem set unique and sorted, return it as an array.
        TODO valid too?"""
        return array(list(set(range(self.subsystems())).intersection(sys)), int)


    def invert_selection(self, sys):
        """Invert and sort a subsystem set."""
        return array(list(set(range(self.subsystems())).difference(sys)), int)


    def fix_phase(self, inplace=False):
        """Apply a global phase convention to a ket state.

        Returns a copy of the state. Additionally, if the state is represented with
        a ket, the copy has a global phase such that the first nonzero element in the
        state vector is real and positive.
        """
        # Ville Bergholm 2009-2011

        s = self._inplacer(inplace)
        if s.is_ket():
            # apply the phase convention: first nonzero element in state vector is real, positive
            v = s.data
            for k in range(v.size):
                temp = v.flat[k]
                if abs(temp) > tol:
                    phase = temp / abs(temp)
                    v /= phase
                    break
        return s


    def normalize(self, inplace=False):
        """Normalize the state to unity."""
        s = self._inplacer(inplace)
        if s.is_ket():
            s.data /= norm(s.data)
        else:
            s.data /= trace(s.data)
        return s


    def purity(self):
        r"""Purity of the state.

        Returns the purity of a normalized state, :math:`p = \mathrm{Tr}(\rho^2)`.
        Equivalent to linear entropy, :math:`S_l = 1-p`.
        """
        # Ville Bergholm 2008-2011

        if self.is_ket():
            return 1
        else:
            # rho is hermitian so purity should be real
            return trace(dot(self.data, self.data)) # .real


    def to_ket(self, inplace=False):
        """Convert the state representation into a ket (if possible).

        If the state is pure returns q, a copy of the state, for which the
        internal representation (q.data) is guaranteed to be a ket vector.
        """
        # Ville Bergholm 2009-2010

        s = self._inplacer(inplace)
        if not s.is_ket():
            # state op
            if abs(s.purity() - 1) > tol:
                raise ValueError('The state is not pure, and thus cannot be represented by a ket vector.')

            d, v = eighsort(s.data)
            s.data = v[:, [0]]  # corresponds to the highest eigenvalue, i.e. 1
            s.fix_phase(inplace = True)  # clean up global phase
            s.dim = (s.dim[0], (1,))

        return s


    def to_op(self, inplace=False):
        """Convert state representation into a state operator.

        Returns q, a copy of the state for which the internal representation 
        (q.data) is guaranteed to be a state operator.
        """
        # Ville Bergholm 2009-2010
        # slight inefficiency when self.is_ket() and inplace==False: the ket data is copied for no reason
        s = self._inplacer(inplace)
        if s.is_ket():
            s.data = np.outer(s.data, s.data.conj())
            s.dim = (s.dim[0], s.dim[0])

        return s


    def trace(self):
        """Trace of the state operator.

        Returns the trace of the state operator of quantum state s.
        For a pure state this is equal to the squared norm of the state vector.
        """
        # Ville Bergholm 2008-2012

        if self.is_ket():
            # squared norm, thus always real
            return vdot(self.data, self.data).real
        else:
            return trace(self.data)  # .real


    def ptrace(self, sys, inplace=False):
        """Partial trace.

        Returns the partial trace of the state
        over the subsystems listed in the vector sys.
        """
        # Ville Bergholm 2008-2010

        s = self.to_op(inplace)
        dim = s.dims()
        n = s.subsystems()
        sys = s.clean_selection(sys)
        keep = s.invert_selection(sys)

        def tensorsum(a, b):
            """Like tensor product, but with sum."""
            #c = log(kron(exp(a), exp(b))) # a perverted way of doing it, the exp overflows...
            c = []
            for k in a:
                c.extend(k + b)
            return array(c)

        # big-endian (C) data ordering
        # we trace over the subsystems in order, starting from the first one
        # partial trace over single system j, performed for every j in sys
        d = list(dim)
        for j in sys:
            muls = index_muls(d)  # muls == [d_{n-1}*...*d_1, d_{n-1}*...*d_2, ..., d_{n-1}, 1]

            # build the index "stencil"
            inds = array([0])
            for k in range(n):
                if k != j:
                    inds = tensorsum(inds, r_[0 : muls[k] * d[k] : muls[k]])
                    # np.arange(d[k]) * muls[k]

            stride = muls[j] # stride for moving the stencil while summing
            temp = len(inds)
            res = zeros((temp, temp), complex) # result
            for k in range(d[j]):
                temp = inds + stride * k
                res += s.data[ix_(temp, temp)]

            s.data = res # replace data
            d[j] = 1  # remove traced-over dimension.

        dim = tuple(array(dim)[keep]) # remove traced-over dimensions for good
        if len(dim) == 0:
            dim = (1,) # full trace gives a scalar

        s.dim = (dim, dim)
        return s



    def ptranspose(self, sys, inplace=False):
        """Partial transpose.

        Returns the partial transpose of the state
        wrt. the subsystems listed in the vector sys.
        """
        # Ville Bergholm 2008-2011

        # TODO what about kets? can we do better?
        s = self.to_op(inplace)
        dim = s.dims()
        n = s.subsystems()
        # total dimension
        orig_d = s.data.shape
        # which systems to transpose
        sys = s.clean_selection(sys)

        # swap the transposed dimensions
        perm = np.arange(2 * n)  # identity permutation
        perm[r_[sys, sys + n]] = perm[r_[sys + n, sys]]

        # flat matrix into tensor, partial transpose, back into a flat matrix
        s.data = s.data.reshape(dim + dim).transpose(perm).reshape(orig_d)
        return s


    def reorder(self, perm, inplace=False):
        """Change the relative order of subsystems in a state.
        ::

          reorder([2, 1, 0])    reverse the order of three subsystems
          reorder([2, 5])       swap subsystems 2 and 5

        Reorders the subsystems of the state according to the permutation perm.

        The permutation vector may consist of either exactly two subsystem indices
        (to be swapped), or a full permutation of subsystem indices.
        """
        # Ville Bergholm 2010

        # this is just an adapter for lmap.reorder
        if self.is_ket():
            perm = (perm, None)
        else:
            perm = (perm, perm)
        return super(state, self).reorder(perm, inplace = inplace)


# physics methods

    def ev(self, A):
        """Expectation value of an observable in the state.

        Returns the expectation value of the observable A in the state.
        A has to be Hermitian.
        """
        # Ville Bergholm 2008

        # TODO for diagonal A, self.ev(A) == sum(A * self.prob())
        if self.is_ket():
            # state vector
            x = vdot(self.data, dot(A, self.data))
        else:
            # state operator
            x = trace(dot(A, self.data))
        return x.real # .real for a Hermitian observable and valid state


    def var(self, A):
        """Variance of an observable in the state.

        Returns the variance of the observable A in the state.
        A has to be Hermitian.
        """
        # Ville Bergholm 2009

        return self.ev(A**2) - self.ev(A)**2


    def prob(self):
        """Measurement probabilities of the state in the computational basis.

        Returns a vector of probabilities of finding a system
        in each of the different states of the computational basis.
        """
        # Ville Bergholm 2009

        if self.is_ket():
            temp = self.data.ravel() # into 1D array
            return (temp * temp.conj()).real  # == np.absolute(self.data) ** 2
        else:
            return diag(self.data).real # .real


    def projector(self):
        """Projection operator defined by the state.

        Returns the projection operator P defined by the state.
        """
        # Ville Bergholm 2009-2014
        if abs(self.purity() - 1) > tol:
            raise ValueError('The state is not pure, and thus does not correspond to a projector.')

        s = self.to_op()
        return lmap(s)


    def u_propagate(self, U):
        """Propagate the state using a unitary.

        Propagates the state using the unitary propagator U,
        returns the resulting state.
        """
        # Ville Bergholm 2009-2010

        if isinstance(U, lmap):
            if self.is_ket():
                return state(U * self)
            else:
                return state(U * self * U.ctranspose())
        elif isinstance(U, np.ndarray):
            # U is a matrix, dims do not change. could also construct an lmap here...
            if self.is_ket():
                return state(dot(U, self.data), self.dims())
            else:
                return state(dot(dot(U, self.data), U.conj().transpose()), self.dims())
        else:
            raise TypeError('States can only be propagated using lmaps and arrays.')


    def propagate(self, G, t, out_func=lambda x, h: deepcopy(x), **kwargs):
        r"""Propagate the state continuously in time.
        ::

          propagate(H, t)                     Hamiltonian
          propagate(L, t)                     Liouvillian
          propagate([H, A_1, A_2, ...], t)    Hamiltonian and Lindblad ops

        Propagates the state using the generator G for the time t,
        returns the resulting state.

        The generator G can either be a

        * Hamiltonian H: :math:`\text{out} = \exp(-i H t) |s\rangle` (or :math:`\exp(-i H t) \rho_s \exp(+i H t)`)
        * Liouvillian superoperator L: :math:`\text{out} = \text{inv\_vec}(\exp(L t) \text{vec}(\rho_s))`
        * list consisting of a Hamiltonian followed by Lindblad operators.
        
        For time-dependent cases, G can be a function G(t) which takes a time instance t
        as input and returns the corresponding generator(s).

        If t is a vector of increasing time instances, returns a list
        containing the propagated state for each time given in t.

        Optional parameters:
        out_func: If given, for each time instance t returns out_func(s(t), G(t)).
        Any unrecognized keyword args are passed on to the ODE solver.
        """
        # Ville Bergholm 2008-2011
        # James Whitfield 2009

        s = self._inplacer(False)
        if isscalar(t):
            t = [t]
        t = asarray(t)
        n = len(t) # number of time instances we are interested in
        out = []
        dim = s.data.shape[0]  # system dimension

        if isinstance(G, types.FunctionType):
            # time dependent
            t_dependent = True
            F = G
            H = G(0)
        else:
            # time independent
            t_dependent = False
            H = G

        if isinstance(H, np.ndarray):
            # matrix
            dim_H = H.shape[1]
            if dim_H == dim:
                gen = 'H'  # Hamiltonian
            elif dim_H == dim ** 2:
                gen = 'L'  # Liouvillian
                s.to_op(inplace = True)
            else:
                raise ValueError('Dimension of the generator does not match the dimension of the state.')
        elif isinstance(H, list):
            # list: Hamiltonian and the Lindblad operators
            dim_H = H[0].shape[1]
            if dim_H == dim:
                gen = 'A'
                s.to_op(inplace = True)

                # HACK, in this case we use an ODE solver anyway
                if not t_dependent:
                    t_dependent = True 
                    F = lambda t: H  # ops stay constant
            else:
                raise ValueError('Dimension of the Lindblad ops does not match the dimension of the state.')
        else:
            raise ValueError("""The second parameter has to be either a matrix, a list,
                             or a function that returns a matrix or a list.""")

        dim = s.data.shape  # may have been switched to operator representation

        if t_dependent:
            # time dependent case, use ODE solver

            # derivative functions for the solver TODO vectorization?
            # H, ket
            def pure_fun(t, y, F):
                "Derivative of a pure state, Hamiltonian."
                return -1j * dot(F(t), y)
            def pure_jac(t, y, F):
                "Jacobian of a pure state, Hamiltonian."
                return -1j * F(t)

            # H, state op
            def mixed_fun(t, y, F):
                "Derivative of a mixed state, Hamiltonian."
                H = -1j * F(t)
                if y.ndim == 1:
                    rho = inv_vec(y, dim)  # into a matrix
                    return vec(dot(H, rho) - dot(rho, H)) # back into a vector
                else:
                    # vectorization, rows of y 
                    d = empty(y.shape, complex)
                    for k in range(len(y)):
                        rho = inv_vec(y[k], dim) # into a matrix
                        d[k] = vec(dot(H, rho) - dot(rho, H)) # back into a vector
                    return d

            # L, state op, same as the H/ket ones, only without the -1j
            def liouvillian_fun(t, y, F):
                "Derivative of a state, Liouvillian."
                return dot(F(t), y)
            def liouvillian_jac(t, y, F):
                "Jacobian of a state, Liouvillian."
                return F(t)

            # A, state op
            def lindblad_fun(t, y, F):
                "Derivative of a mixed state, Lindbladian."
                X = F(t)  # X == [H, A_1, A_2, ..., A_n]
                H = -1j * X[0] # -1j * Hamiltonian
                Lind = X[1:]   # Lindblad ops
                if y.ndim == 1:
                    rho = inv_vec(y, dim)  # into a matrix
                    temp = dot(H, rho) - dot(rho, H)
                    for A in Lind:
                        ac = 0.5 * dot(A.conj().transpose(), A)
                        temp += dot(dot(A, rho), A.conj().transpose()) -dot(ac, rho) -dot(rho, ac)
                    return vec(temp) # back into a vector
                else:
                    # vectorization, rows of y
                    d = empty(y.shape, complex)
                    for k in range(len(y)):
                        rho = inv_vec(y[k], dim)  # into a matrix
                        temp = dot(H, rho) - dot(rho, H)
                        for A in Lind:
                            ac = 0.5 * dot(A.conj().transpose(), A)
                            temp += dot(dot(A, rho), A.conj().transpose()) -dot(ac, rho) -dot(rho, ac)
                        d[k] = vec(temp)  # back into a vector
                    return d

            # what kind of generator are we using?
            if gen == 'H':  # Hamiltonian
                if dim[1] == 1:
                    func, jac = pure_fun, pure_jac
                else:
                    func, jac = mixed_fun, None
            elif gen == 'L':  # Liouvillian
                func, jac = liouvillian_fun, liouvillian_jac
            else: # 'A'  # Hamiltonian and Lindblad operators in a list
                func, jac = lindblad_fun, None

            # do we want the initial state too? (the integrator can't handle t=0!)
            if t[0] == 0:
                out.append(out_func(s, F(0)))
                t = t[1:]

            # ODE solver default parameters
            odeopts = {'rtol' : 1e-4,
                       'atol' : 1e-6,
                       'method' : 'bdf', # 'adams' for non-stiff cases
                       'with_jacobian' : True}
            odeopts.update(kwargs) # user options

            # run the solver
            r = ode(func, jac).set_integrator('zvode', **odeopts)
            r.set_initial_value(vec(s.data), 0.0).set_f_params(F).set_jac_params(F)
            for k in t:
                r.integrate(k)  # times must be in increasing order, NOT include zero(!)
                if not r.successful():
                    raise RuntimeError('ODE integrator failed.')
                s.data = inv_vec(r.y, dim)
                out.append(out_func(s, F(k)))

        else:
            # time independent case
            if gen == 'H':
                if dim_H < 500:
                    # eigendecomposition
                    d, v = eigh(H)
                    for k in t:
                        # propagator
                        U = dot(dot(v, diag(exp(-1j * k * d))), v.conj().transpose())
                        out.append(out_func(s.u_propagate(U), H))
                else:
                    # Krylov subspace method
                    # FIXME imaginary time doesn't yet work
                    w, err, hump = expv(-1j * t, H, s.data)
                    for k in range(n):
                        s.data = w[k, :]  # TODO state ops
                        out.append(out_func(s, H))
            elif gen == 'L':
                # Krylov subspace method
                w, err, hump = expv(t, H, vec(s.data))
                for k in range(n):
                    s.data = inv_vec(w[k, :])
                    out.append(out_func(s, H))

        if len(out) == 1:
            return out[0] # don't bother to wrap a single output in a list
        else:
            return out


    def kraus_propagate(self, E):
        r"""Apply a quantum operation to the state.

        Applies the quantum operation E to the state.
        :math:`E = [E_1, E_2, \ldots]` is a set of Kraus operators.
        """
        # Ville Bergholm 2009

        # TODO allow the user to apply E only to some subsystems of s0
        n = len(E)
        # TODO: If n > prod(dims(s))^2, there is a simpler equivalent
        # operation. Should the user be notified?
        def test_kraus(E):
            "Check if E represents a physical quantum operation."
            temp = 0
            for k in E:
                temp += dot(k.conj().transpose(), k)
            if norm(temp.data - np.eye(temp.shape)) > tol:
                _warn('Unphysical quantum operation.')

        if self.is_ket():
            if n == 1:
                return self.u_propagate(E[0]) # remains a pure state

        s = self.to_op()
        q = state(zeros(s.data.shape, complex), s.dims())
        for k in E:
            q += s.u_propagate(k)
        return q


    def measure(self, M=None, do='R'):
        r"""Quantum measurement.
        ::

          (p, res, c)
            = measure()                 measure the entire system projectively
            = measure((1, 4))           measure subsystems 1 and 4 projectively
            = measure([M_1, M_2, ...])  perform a general measurement
            = measure(A)                measure a Hermitian observable A

        Performs a quantum measurement on the state.

        * If no M is given, a full projective measurement in the
          computational basis is performed.

        * If a list/tuple of subsystems is given as the second parameter, only
          those subsystems are measured, projectively, in the
          computational basis.

        * A general measurement may be performed by giving a complete set
          of measurement operators :math:`[M_1, M_2, \ldots]` as the second parameter.
          A POVM can be emulated using :math:`M_i = \sqrt{P_i}` and discarding the collapsed state.

        * Finally, if the second parameter is a single Hermitian matrix A, the
          corresponding observable is measured. In this case the second
          column of p contains the eigenvalue of A corresponding to each
          measurement result.

        p = measure(..., do='P') returns the vector p, where p[k] is the probability of
        obtaining result k in the measurement. For a projective measurement
        in the computational basis this corresponds to the ket :math:`|k\rangle`.

        (p, res) = measure(...) additionally returns the index of the result of the
        measurement, res, chosen at random from the probability distribution p.
 
        (p, res, c) = measure(..., do='C') additionally gives c, the collapsed state
        corresponding to the measurement result res.
        """
        # Ville Bergholm 2009-2010

        def rand_measure(p):
            """Result of a random measurement using the prob. distribution p."""
            return nonzero(rand() <= cumsum(p))[0][0]

        perform = True
        collapse = False
        do = do.upper()
        if do in ('C', 'D'):
            collapse = True
        elif do == 'P':
            perform = False

        d = self.dims()

        if M == None:
            # full measurement in the computational basis
            p = self.prob()  # probabilities 
            if perform:
                res = rand_measure(p)
                if collapse:
                    s = state(res, d) # collapsed state

        elif isinstance(M, np.ndarray):
            # M is a matrix TODO lmap?
            # measure the given Hermitian observable
            a, P = spectral_decomposition(M)
            m = len(a)  # number of possible results

            p = zeros((m, 2))
            for k in range(m):
                p[k, 0] = self.ev(P[k])  # probabilities
            p[:, 1] = a  # corresponding measurement results

            if perform:
                res = rand_measure(p)
                if collapse:
                    # collapsed state
                    ppp = P[res]  # Hermitian projector
                    s = deepcopy(self)
                    if self.is_ket():
                        s.data = dot(ppp, s.data) / sqrt(p[res, 0])
                    else:
                        s.data = dot(dot(ppp, s.data), ppp) / p[res, 0]

        elif isinstance(M, (list, tuple)):
            if isinstance(M[0], numbers.Number):
                # measure a set of subsystems in the computational basis
                sys = self.clean_selection(M)
                d = array(d)

                # dimensions of selected subsystems and identity ops between them
                # TODO sequential measured subsystems could be concatenated as well
                q = len(sys)
                pdims = []
                start = 0  # first sys not yet included
                for k in sys:
                    pdims.append(prod(d[start:k])) # identity
                    pdims.append(d[k]) # selected subsys
                    start = k+1

                pdims.append(prod(d[start:])) # last identity

                # index multipliers 
                muls = index_muls(d[sys])
                # now muls == [..., d_s{q-1}*d_s{q}, d_s{q}, 1]

                m = muls[0] * d[sys][0] # number of possible results == prod(d[sys])

                def build_stencil(j, q, pdims, muls):
                    """Projector to state j (diagonal because we project into the computational basis)"""
                    stencil = ones(pdims[0]) # first identity
                    for k in range(q):
                        # projector for system k
                        temp = zeros(pdims[2*k + 1])
                        temp[int(j / muls[k]) % pdims[2*k + 1]] = 1
                        stencil = kron(kron(stencil, temp), ones(pdims[2*k + 2])) # temp + next identity
                    return stencil

                # sum the probabilities
                p = zeros(m)
                born = self.prob()
                for j in range(m):
                    p[j] = dot(build_stencil(j, q, pdims, muls), born)

                if perform:
                    res = rand_measure(p)
                    if collapse:
                        # collapsed state
                        s = deepcopy(self)
                        R = build_stencil(res, q, pdims, muls) # diagonal of a diagonal projector (just zeros and ones)

                        if do == 'D':
                            # discard the measured subsystems from s
                            d = np.delete(d, sys)
                            keep = (R == 1)  # indices of elements to keep
        
                            if self.is_ket():
                                s.data = s.data[keep] / sqrt(p[res])
                            else:
                                s.data = s.data[:, keep][keep, :] / p[res]

                            s = state(s.data, d)
                        else:
                            if self.is_ket():
                                s.data = R.reshape(-1, 1) * s.data / sqrt(p[res]) # collapsed state
                            else:
                                s.data = np.outer(R, R) * s.data / p[res] # collapsed state, HACK
            else:
                # otherwise use set M of measurement operators (assumed complete!)
                m = len(M)

                # probabilities
                p = zeros(m)
                for k in range(m):
                    p[k] = self.ev(dot(M[k].conj().transpose(), M[k]))  #  M^\dagger M  is Hermitian
                    # TODO for kets, this is slightly faster:
                    #temp = dot(M[k], self.data)
                    #p[k] = vdot(temp, temp)

                if perform:
                    res = rand_measure(p)
                    if collapse:
                        s = deepcopy(self)
                        if self.is_ket():
                            s.data = dot(M[res], s.data) / sqrt(p[res])
                        else:
                            s.data = dot(dot(M[res], s.data), M[res].conj().transpose()) / p[res]
        else:
            raise ValueError('Unsupported input type.')
        if collapse:
            return p, res, s
        elif perform:
            return p, res
        else:
            return p



# quantum information methods

    def fidelity(self, r):
        r"""Fidelity of two states.

        Fidelity of two state operators \rho and \sigma is defined as
        :math:`F(\rho, \sigma) = \mathrm{Tr} \sqrt{\sqrt{\rho} \sigma \sqrt{\rho}}`.
        For state vectors this is equivalent to the overlap, :math:`F = |\langle a|b\rangle|`.

        Fidelity is symmetric in its arguments and bounded in the interval [0,1].
        See :cite:`NC`, chapter 9.2.2.
        """
        # Ville Bergholm 2009-2010

        if not isinstance(r, state):
            raise TypeError('Not a state.')

        if self.is_ket():
            if r.is_ket():
                return abs(vdot(self.data, r.data))
            else:
                return sqrt(vdot(self.data, dot(r.data, self.data)).real)  # .real
        else:
            if r.is_ket():
                return sqrt(vdot(r.data, dot(self.data, r.data)).real)  # .real
            else:
                temp = sqrtm(self.data)
                return trace(sqrtm(dot(dot(temp, r.data), temp))).real  # .real


    def trace_dist(self, r):
        r"""Trace distance of two states.

        Trace distance between state operators r and s is defined as
        :math:`D(r, s) = \frac{1}{2} \mathrm{Tr}(\sqrt{A^\dagger A})`, where A = r-s.

        Equivalently :math:`D(r, s) = \frac{1}{2} \sum_k |\lambda_k|`, where :math:`\lambda_k`
        are the eigenvalues of A (since A is Hermitian). See :cite:`NC`, chapter 9.2.1.
        """
        # Ville Bergholm 2009

        if not isinstance(r, state):
            raise TypeError('Not a state.')

        # avoid copying state ops since we just do read-only stuff here
        S = self.to_op() if self.is_ket() else self
        R = r.to_op() if r.is_ket() else r

        A = R.data - S.data
        return 0.5 * sum(abs(eigvalsh(A)))
        #return 0.5*trace(sqrtm(A'*A))



    def schmidt(self, sys=None, full=False):
        r"""Schmidt decomposition.
        ::

          lambda       = schmidt(sys)
          lambda, u, v = schmidt(sys, full=True)

        Calculates the Schmidt decomposition of the (pure) state.
        Subsystems listed in vector sys constitute part A, the rest forming part B.
        Vector lambda will contain the Schmidt coefficients.

        If required, matrices u and v will contain the corresponding orthonormal
        Schmidt bases for A and B, respectively, as column vectors, i.e.

        .. math::

           |k\rangle_A = u[:, k],  \qquad |k\rangle_B = v[:, k].

        The state is then given by :math:`\sum_k \lambda_k |k\rangle_A \otimes |k\rangle_B`.

        See :cite:`NC`, chapter 2.5.
        """
        # Ville Bergholm 2009-2014

        dim = array(self.dims())
        n = self.subsystems()

        if sys == None:
            if n == 2:
                # reasonable choice
                sys = (0,)
            else:
                raise ValueError('Requires a state and a vector of subsystems.')

        try:
            s = self.to_ket()
        except ValueError:
            raise ValueError('Schmidt decomposition is only defined for pure states.')

        # complement of sys, dimensions of the partitions
        sys = s.clean_selection(sys)
        compl = s.invert_selection(sys)
        d1 = prod(dim[sys])
        d2 = prod(dim[compl])
        perm = r_[sys, compl]

        if all(perm == range(n)):
            # nothing to do
            pass
        else:
            # reorder the system according to the partitioning
            s.reorder(perm, inplace = True)

        # order the coefficients into a matrix, take an svd
        if not full:
            return svdvals(s.data.reshape(d1, d2))
        else:
            u, s, vh = svd(s.data.reshape(d1, d2), full_matrices = False)
            # note the definition of vh in svd
            return s, u, vh.transpose()



    def entropy(self, sys=None, alpha=1):
        r"""Von Neumann or Renyi entropy of the state.

        Returns the Renyi entropy of order :math:`\alpha`,
        :math:`S_\alpha(\rho) = \frac{1}{1-\alpha} \log_2 \mathrm{Tr}(\rho^\alpha)`.
        
        When :math:`\alpha = 1`, this coincides with the von Neumann entropy
        :math:`S(\rho) = -\mathrm{Tr}(\rho \log_2(\rho))`.

        If sys == None, returns the entropy of the state.
        If sys is a vector of subsystem indices, returns the
        entropy of entanglement of the state wrt. the partitioning
        defined by sys. Entropy of entanglement is only defined for pure states.
        """
        # Ville Bergholm 2009-2012

        if sys != None:
            s = self.to_ket().ptrace(sys) # partial trace over one partition
        else:
            s = self

        if s.is_ket():
            return 0

        p = eigvalsh(s.data)
        if alpha != 1:
            # Rényi entropy
            return np.log2(sum(p ** alpha)) / (1 - alpha)
        else:
            # Von Neumann entropy
            p[p == 0] = 1   # avoid trouble with the logarithm
            return -dot(p, np.log2(p))


    def concurrence(self, sys=None):
        """Concurrence of the state.

        Returns the concurrence of the state s wrt. the partitioning
        given by the listing of subsystems in the vector sys.

        See :cite:`Wootters`, :cite:`Horodecki`.
        """
        # Ville Bergholm 2006-2014
        # TODO rewrite, check
        if abs(self.trace() - 1) > tol:
            _warn('State not properly normalized.')

        dim = self.dims()

        if sys != None:
            # concurrence between a qubit and a larger system
            if not(len(sys) == 1 and dim[sys] == 2):
                raise ValueError('Concurrence only defined between a qubit and another system.')

            if abs(self.purity() - 1) > tol:
                raise ValueError('Not a pure state.')

            # pure state
            #n = len(dim)
            rho_A = self.ptrace(self.invert_selection(sys)) # trace over everything but sys
            return 2 * sqrt(det(rho_A.data).real) # = sqrt(2*(1-real(trace(temp*temp)))), .real

        # concurrence between two qubits
        if self.subsystems() != 2 or any(dim != array([2, 2])):
            # not a two-qubit state
            raise ValueError('Not a two-qubit state.')

        X = kron(sy, sy)
        p = self.data
        if self.is_ket():
            # ket
            return abs(dot(dot(p.transpose(), X), p))

            # find the coefficients a of the state ket in the magic base
            # phi+-, psi+-,  = triplet,singlet
            #bell = [1 i 0 0 0 0 i 1 0 0 i -1 1 -i 0 0]/sqrt(2)
            #a = bell'*p
            #C = abs(sum(a ** 2))
        else:
            # state operator
            temp = dot(p, X)  # X.conj() == X so this works
            temp = dot(temp, temp.conj())  # == p * X * conj(p) * X
            if abs(self.purity() - 1) > tol:
                L = sqrt(sort(np.linalg.eigvals(temp).real)[::-1]).real  # .real?
                return max(0, L[1] -L[2] -L[3] -L[4])
            else:
                return sqrt(trace(temp).real) # same formula as for state vecs, .real?


    def negativity(self, sys):
        """Negativity of the state.

        Returns the negativity of the state wrt. the partitioning
        given by the listing of subsystems in the vector sys.

        See :cite:`Peres`, :cite:`Horodecki1`
        """
        # Ville Bergholm 2008-2014

        s = self.ptranspose(sys)  # partial transpose the state
        x = svdvals(s.data)  # singular values
        return (sum(x) - 1) / 2


    def lognegativity(self, sys):
        """Logarithmic negativity of the state.

        Returns the logarithmic negativity of the state wrt. the partitioning
        given by the listing of subsystems in the vector sys.
        """
        # Ville Bergholm 2008-2014

        return np.log2(2 * self.negativity(sys) + 1)


    def scott(self, m):
        """Scott's average bipartite entanglement measure.

        Returns the vector Q containing the terms of the Scott entanglement measure
        of the system for partition size m.

        When m = 1 this is coincides with the Meyer-Wallach entanglement measure.

        See :cite:`Love`, :cite:`Scott`, :cite:`MW`.
        """
        # Jacob D. Biamonte 2008
        # Ville Bergholm 2008-2014

        import itertools

        dim = self.dims()
        n = self.subsystems()

        if m < 1 or m > n-1:
            raise ValueError('Partition size must be between 1 and n-1.')

        D = min(dim) # FIXME correct for arbitrary combinations of qudits??
        N = sp.misc.comb(n, m, exact=True)
        C = (D**m / (D**m - 1)) / N  # normalization

        Q = empty((N,))
        # Loop over all m-combinations of n subsystems, trace over everything except them.
        # reversed() fixes the order since we are actually looping over the complements.
        for k, sys in enumerate(reversed(list(itertools.combinations(range(n), n-m)))):
            temp = self.ptrace(sys)  # trace over everything except S_k
            # NOTE: For pure states, tr(\rho_S^2) == tr(\rho_{\bar{S}}^2)
            temp = 1 - trace(np.linalg.matrix_power(temp.data, 2))
            Q[k] = C * temp.real
        return Q



    def locc_convertible(self, t, sys):
        """LOCC convertibility of states.

        For bipartite pure states s and t, returns True if s can be converted to t
        using local operations and classical communication (LOCC).
        sys is a vector of subsystems defining the partition.

        See :cite:`NC`, chapter 12.5.1
        """
        # Ville Bergholm 2010

        if not equal_dims(self, t):
            raise ValueError('States must have equal dimensions.')

        try:
            s = self.to_ket()
            t = t.to_ket()
        except ValueError:
            raise ValueError('Not implemented for nonpure states.')

        s.ptrace(sys, inplace = True)
        t.ptrace(sys, inplace = True)
        return majorize(eigvalsh(s.data), eigvalsh(t.data))



    def plot(self, symbols=3):
        """State tomography plot.

        Plots the probabilities of finding a system in this state
        in the different computational basis states upon measurement.
        Relative phases are represented by the colors of the bars.

        If the state is nonpure, also plots the coherences.
        """
        # Ville Bergholm 2009-2010

        import matplotlib.pyplot as plt
        from matplotlib import cm, colors

        dim = self.dims()
        n = self.subsystems()

        # prepare labels
        m = min(n, symbols)  # at most three symbols
        d = dim[:m]
        nd = prod(d)
        rest = '0' * (n-m) # the rest is all zeros
        ticklabels = []
        for k in range(nd):
            temp = array_to_numstr(np.unravel_index(k, d))
            ticklabels.append(temp + rest)

        ntot = prod(dim)
        skip = ntot / nd  # only every skip'th state gets a label to avoid clutter
        ticks = r_[0 : ntot : skip]

        N = self.data.shape[0]
        fig = plt.gcf()
        fig.clf() # clear the figure

        # color normalization
        nn = colors.Normalize(vmin = -1, vmax = 1, clip = True)
        def phases(A):
            """Phase normalized to (-1,1]"""
            return np.angle(A) / np.pi

        if self.is_ket():
            s = self.fix_phase()
            c = phases(s.data.ravel())  # use phases as colors
            ax = fig.gca()

            width = 0.8
            bars = ax.bar(range(N), s.prob(), width)
            # color bars using phase data
            colormapper = cm.ScalarMappable(norm=nn, cmap=cm.hsv)
            colormapper.set_array(c)
            for b in range(N):
                bars[b].set_edgecolor('k')
                bars[b].set_facecolor(colormapper.to_rgba(c[b]))

            # add a colorbar
            cb = fig.colorbar(colormapper, ax = ax, ticks = linspace(-1, 1, 5))
            cb.ax.set_yticklabels([r'$-\pi$', r'$-\pi/2$', '0', r'$\pi/2$', r'$\pi$'])

            # the way it should work (using np.broadcast, ScalarMappable)
            #bars = ax.bar(range(N), s.prob(), color=c, cmap=cm.hsv, norm=whatever, align='center')
            #cb = fig.colorbar(bars, ax = ax, ticks = linspace(-1, 1, 5))

            ax.set_xlabel('Basis state')
            ax.set_ylabel('Probability')
            ax.set_xticks(ticks + width / 2) # shift by half the bar width
            ax.set_xticklabels(ticklabels)
        else:
            import mpl_toolkits.mplot3d

            c = phases(self.data)  # use phases as colors
            ax = fig.add_subplot(111, projection = '3d')
            ax.view_init(40, -115)

            width = 0.6  # bar width
            temp = np.arange(-width/2, N-1) # center the labels
            x, y = meshgrid(temp, temp[::-1])
            x = x.ravel()
            y = y.ravel()
            z = abs(self.data.ravel())
            dx = width * ones(x.shape)
            ax.bar3d(x, y, zeros(x.shape), dx, dx, z, color='b')

            # now the colors
            pcol = ax.get_children()[3]  # FIXME we need a robust way of getting the poly3Dcollection
            pcol.set_norm(nn)
            pcol.set_cmap(cm.hsv)
            pcol.set_array(kron(c.ravel(), (1,)*6))  # six faces per bar

            # add a colorbar
            cb = fig.colorbar(pcol, ax = ax, ticks = linspace(-1, 1, 5))
            cb.ax.set_yticklabels([r'$-\pi$', r'$-\pi/2$', '0', r'$\pi/2$', r'$\pi$'])

            # the way it should work (using np.broadcast, ScalarMappable)
            #x, y = meshgrid(temp, temp)
            #pcol = ax.bar3d(x, y, 0, width, width, abs(self.data), color=c, cmap=cm.hsv, norm=whatever, align='center')

            ax.set_xlabel('Col state')
            ax.set_ylabel('Row state')
            ax.set_zlabel('$|\\rho|$')
            #ax.set_xticks(ticks)
            ax.set_xticklabels(ticklabels)
            #ax.set_yticks(ticks)
            ax.set_yticklabels(ticklabels[::-1])
            #ax.set_alpha(0.8)
            # TODO ticks, ticklabels, alpha

        plt.show()
        return ax


# other state representations

    def bloch_vector(self):
        r"""Generalized Bloch vector.

        Returns the generalized Bloch vector A corresponding to the state.

        For an n-subsystem state the generalized Bloch vector is an order-n correlation
        tensor defined in terms of the standard Hermitian tensor basis B
        corresponding to state dimensions:

        .. math::

           A_{ijk\ldots} = \sqrt{D} \mathrm{Tr}(\rho_s  B_{ijk\ldots}),

        where D = prod(self.dims()). A is always real since :math:`\rho_s` is Hermitian.
        For valid, normalized states

           self.purity() <= 1   implies   norm(A, 'fro')  <= sqrt(D)

           self.trace()   = 1   implies   A[0, 0, ..., 0] == 1

        E.g. for a single-qubit system norm(A, 'fro') <= sqrt(2).
        """
        # Ville Bergholm 2009-2011

        dim = self.dims()
        G = tensorbasis(dim)
        a = []
        for g in G:
            a.append(self.ev(g))
        a = array(a) * sqrt(prod(dim)) # to match the usual Bloch vector normalization

        # into an array, one dimension per subsystem
        return a.reshape(array(dim) ** 2)




    def tensor(*arg):
        """Tensor product of states.

        Returns the tensor product state of states s1, s2, ...
        """
        # Ville Bergholm 2009-2010

        # if all states are kets, keep the result state a ket
        pure = True
        for k in arg:
            if not k.is_ket():
                pure = False
                break

        if not pure:
            # otherwise convert all states to state ops before tensoring
            temp = []
            for k in arg:
                temp.append(k.to_op())
            arg = temp

        return state(lmap_tensor(*arg))


    @staticmethod
    def bloch_state(A, dim=None):
        r"""State corresponding to a generalized Bloch vector.
        s = bloch_state(A)       assume 
        s = bloch_state(A, dim)  give state dimensions explicitly

        Returns the state s corresponding to the generalized Bloch vector A.

        A is defined in terms of the standard Hermitian tensor basis B
        corresponding to the dimension vector dim. If dim is not given, it is assumed
        to be sqrt(A.shape).

        .. math::

           \rho_s = \sum_{ijk\ldots} A_{ijk\ldots} B_{ijk\ldots} / \sqrt{D},

        where D = prod(dim). For valid states norm(A, 2) <= sqrt(D).
        """
        # Ville Bergholm 2009-2011

        if dim == None:
            dim = tuple(sqrt(A.shape).astype(int))  # s == dim ** 2

        G = tensorbasis(dim)
        d = prod(dim)
        rho = zeros((d, d), complex)
        for k, a in enumerate(A.flat):
            rho += a * G[k]

        C = 1/sqrt(d) # to match the usual Bloch vector normalization
        return state(C * rho, dim)



# wrappers

def fidelity(s, t):
    """Wrapper for state.fidelity."""
    return s.fidelity(t)


def trace_dist(s, t):
    """Wrapper for state.trace_dist."""
    return s.trace_dist(t)



# -*- coding: utf-8 -*-
"""Basic definitions."""
# Ville Bergholm 2008-2014

from __future__ import division, absolute_import, print_function, unicode_literals

from numpy import array, eye, sqrt, finfo

__all__ = ['I', 'sx', 'sy', 'sz', 'p0', 'p1', 'H',
           'Q_Bell', 'tol']


# Pauli matrices
I  = eye(2)
sx = array([[0, 1], [1, 0]])
sy = array([[0, -1j], [1j, 0]])
sz = array([[1, 0], [0, -1]])

# qubit projectors
p0 = array([[1, 0], [0, 0]])
p1 = array([[0, 0], [0, 1]])

# easy Hadamard
H = array([[1, 1], [1, -1]]) / sqrt(2)

# magic basis
Q_Bell = array([[1, 0, 0, 1j], [0, 1j, 1, 0], [0, 1j, -1, 0], [1, 0, 0, -1j]]) / sqrt(2)


# error tolerance
tol = max(1e-8, finfo(float).eps)



# -*- coding: utf-8 -*-
# Author: Ville Bergholm 2011-2014
r"""
Harmonic oscillators (:mod:`qit.ho`)
====================================

.. currentmodule:: qit.ho

This module simulates harmonic oscillators by truncating the state
space dimension to a finite value. Higher truncation limits give more accurate results.
All the functions in this module operate in the truncated number basis
:math:`\{|0\rangle, |1\rangle, ..., |n-1\rangle\}`
of the harmonic oscillator, where n is the truncation dimension.

The corresponding truncated annihilation operator can be obtained with :func:`qit.utils.boson_ladder`.


Contents
--------

.. autosummary::

   coherent_state
   position_state
   momentum_state
   position
   momentum
   displace
   squeeze
   husimi
   wigner
"""

from __future__ import division, absolute_import, print_function, unicode_literals

from numpy import array, mat, empty, arange, diag, sqrt, ones, prod, sqrt, pi, isscalar, linspace, newaxis
from scipy.misc import factorial
from scipy.linalg import expm, norm

from .base import tol
from .state import state
from .utils import boson_ladder, comm

__all__ = ['coherent_state', 'displace', 'squeeze', 'position', 'momentum',
           'position_state', 'momentum_state', 'husimi', 'wigner']


# default truncation limit for number states
default_n = 30

def coherent_state(alpha, n=default_n):
    r"""Coherent states of a harmonic oscillator.

    Returns the n-dimensional approximation to the
    coherent state :math:`|\alpha\rangle`,

    .. math::

       |\alpha\rangle := D(\alpha) |0\rangle
       = e^{-\frac{|\alpha|^2}{2}} \sum_{k=0}^\infty \frac{\alpha^k}{\sqrt{k!}} |k\rangle,

    in the number basis. :math:`a|\alpha\rangle = \alpha |\alpha\rangle`.
    """
    # Ville Bergholm 2010

    k = arange(n)
    ket = (alpha ** k) / sqrt(factorial(k))
    return state(ket, n).normalize()
    #s = state(0, n).u_propagate(expm(alpha * mat(boson_ladder(n)).H))
    #s = state(0, n).u_propagate(displace(alpha, n))
    #s *= exp(-abs(alpha) ** 2 / 2) # normalization


def displace(alpha, n=default_n):
    r"""Bosonic displacement operator.

    Returns the n-dimensional approximation for the bosonic
    displacement operator

    .. math::

       D(\alpha) := \exp\left(\alpha a^\dagger - \alpha^* a\right)
       = \exp\left( i \sqrt{2} \left(Q \mathrm{Im}(\alpha) -P \mathrm{Re}(\alpha)\right)\right)

    in the number basis. This yields

    .. math::

       D(\alpha) Q D^\dagger(\alpha) &= Q -\sqrt{2} \textrm{Re}(\alpha) \mathbb{I},\\
       D(\alpha) P D^\dagger(\alpha) &= P -\sqrt{2} \textrm{Im}(\alpha) \mathbb{I},

    and thus the displacement operator displaces the state of a harmonic oscillator in phase space.
    """
    # Ville Bergholm 2010

    if not isscalar(alpha):
        raise TypeError('alpha must be a scalar.')

    a = mat(boson_ladder(n))
    return array(expm(alpha * a.H -alpha.conjugate() * a))


def squeeze(z, n=default_n):
    r"""Bosonic squeezing operator.

    Returns the n-dimensional approximation for the bosonic
    squeezing operator

    .. math::

       S(z) := \exp\left(\frac{1}{2} (z^* a^2 - z a^{\dagger 2})\right)
       = \exp\left(\frac{i}{2} \left((QP+PQ)\mathrm{Re}(z) +(P^2-Q^2)\mathrm{Im}(z)\right)\right)

    in the number basis.
    """
    # Ville Bergholm 2010
    if not isscalar(z):
        raise TypeError('z must be a scalar.')

    a = mat(boson_ladder(n))
    return array(expm(0.5 * (z.conjugate() * (a ** 2) - z * (a.H ** 2))))


def position(n=default_n):
    r"""Position operator.

    Returns the n-dimensional approximation of the
    dimensionless position operator Q in the number basis.

    .. math::

       Q &= \sqrt{\frac{m \omega}{\hbar}}   q =    (a+a^\dagger) / \sqrt{2},\\
       P &= \sqrt{\frac{1}{m \hbar \omega}} p = -i (a-a^\dagger) / \sqrt{2}.

    (Equivalently, :math:`a = (Q + iP) / \sqrt{2}`).
    These operators fulfill :math:`[q, p] = i \hbar, \quad  [Q, P] = i`.
    The Hamiltonian of the harmonic oscillator is

    .. math::

       H = \frac{p^2}{2m} +\frac{1}{2} m \omega^2 q^2
         = \frac{1}{2} \hbar \omega (P^2 +Q^2)
         = \hbar \omega (a^\dagger a +\frac{1}{2}).
    """
    # Ville Bergholm 2010

    a = mat(boson_ladder(n))
    return array(a + a.H) / sqrt(2)


def momentum(n=default_n):
    """Momentum operator.

    Returns the n-dimensional approximation of the
    dimensionless momentum operator P in the number basis.

    See :func:`position`.
    """
    # Ville Bergholm 2010

    a = mat(boson_ladder(n))
    return -1j*array(a - a.H) / sqrt(2)


def position_state(q, n=default_n):
    r"""Position eigenstates of a harmonic oscillator.

    Returns the n-dimensional approximation of the eigenstate :math:`|q\rangle`
    of the dimensionless position operator Q in the number basis.

    See :func:`position`, :func:`momentum`.

    Difference equation:

    .. math::

       r_1 &= \sqrt{2} q r_0,\\
       \sqrt{k+1} r_{k+1} &= \sqrt{2} q r_k -\sqrt{k} r_{k-1}, \qquad \text{when} \quad k >= 1.
    """
    # Ville Bergholm 2010

    ket = empty(n, dtype=complex)
    temp = sqrt(2) * q
    ket[0] = 1  # arbitrary nonzero initial value r_0
    ket[1] = temp * ket[0]
    for k in range(2, n):
        ket[k] = temp/sqrt(k) * ket[k - 1] -sqrt((k-1) / k) * ket[k - 2]
    ket /= norm(ket)  # normalize
    return state(ket, n)


def momentum_state(p, n=default_n):
    r"""Momentum eigenstates of a harmonic oscillator.

    Returns the n-dimensional approximation of the eigenstate :math:`|p\rangle`
    of the dimensionless momentum operator P in the number basis.

    See :func:`position`, :func:`momentum`.

    Difference equation:

    .. math::

       r_1 &= i \sqrt{2} p r_0,\\
       \sqrt{k+1} r_{k+1} &= i \sqrt{2} p r_k +\sqrt{k} r_{k-1}, \qquad \text{when} \quad k >= 1.
    """
    # Ville Bergholm 2010

    ket = empty(n, dtype=complex)
    temp = 1j * sqrt(2) * p
    ket[0] = 1  # arbitrary nonzero initial value r_0
    ket[1] = temp * ket[0]
    for k in range(2, n):
        ket[k] = temp/sqrt(k) * ket[k - 1] +sqrt((k-1) / k) * ket[k - 2]
    ket /= norm(ket)  # normalize
    return state(ket, n)


def husimi(s, alpha=None, z=0, res=(40, 40), lim=(-2, 2, -2, 2)):
    r"""Husimi probability distribution.
    ::

      H       = husimi(s, alpha[, z=])
      H, a, b = husimi(s, res=xxx, lim=yyy[, z=])

    Returns the Husimi probability distribution
    :math:`H(\mathrm{Im} \alpha, \mathrm{Re} \alpha)` corresponding to the harmonic
    oscillator state s given in the number basis:

    .. math::

       H(s, \alpha, z) = \frac{1}{\pi} \langle\alpha, z| \rho_s |\alpha, z\rangle

    z is the optional squeezing parameter for the reference state:
    :math:`|\alpha, z\rangle := D(\alpha) S(z) |0\rangle`.
    The integral of H is normalized to unity.
    """
    # Ville Bergholm 2010

    if alpha == None:
        # return a 2D grid of W values
        a = linspace(lim[0], lim[1], res[0])
        b = linspace(lim[2], lim[3], res[1])
        #a, b = ogrid[lim[0]:lim[1]:1j*res[0], lim[2]:lim[3]:1j*res[1]]
        alpha = a + 1j*b[:, newaxis]
        return_ab = True
    else:
        return_ab = False

    # reference state
    n = prod(s.dims())
    ref = state(0, n).u_propagate(squeeze(z, n))
    ref /= sqrt(pi) # normalization included for convenience

    H = empty(alpha.shape)
    for k, c in enumerate(alpha.flat):
        temp = ref.u_propagate(displace(c, n))
        H.flat[k] = s.fidelity(temp) ** 2

    if return_ab:
        H = (H, a, b)
    return H


def wigner(s, alpha=None, res=(20, 20), lim=(-2, 2, -2, 2)):
    r"""Wigner quasi-probability distribution.
    ::

      W       = wigner(s, alpha)
      W, a, b = wigner(s, res=xxx, lim=yyy)

    Returns the Wigner quasi-probability distribution
    :math:`W(\mathrm{Im} \alpha, \mathrm{Re} \alpha)` corresponding to the harmonic
    oscillator state s given in the number basis.

    For a normalized state, the integral of W is normalized to unity.

    NOTE: The truncation of the number state space to a finite dimension
    results in spurious circular ripples in the Wigner function outside
    a given radius. To increase the accuracy, increase the state space dimension.
    """
    # Ville Bergholm 2010

    if alpha == None:
        # return a grid of W values for a grid of alphas
        a = linspace(lim[0], lim[1], res[0])
        b = linspace(lim[2], lim[3], res[1])
        #a, b = ogrid[lim[0]:lim[1]:1j*res[0], lim[2]:lim[3]:1j*res[1]]
        alpha = a + 1j*b[:, newaxis]
        return_ab = True
    else:
        return_ab = False

    # parity operator (diagonal)
    n = prod(s.dims())
    P = ones(n)
    P[1:n:2] = -1
    P *= 2 / pi  # include Wigner normalization here for convenience

    W = empty(alpha.shape)
    for k, c in enumerate(alpha.flat):
        temp = s.u_propagate(displace(-c, n))
        W.flat[k] = sum(P * temp.prob().real) # == ev(temp, P).real

    if return_ab:
        W = (W, a, b)
    return W



# -*- coding: utf-8 -*-
"Unit tests for qit.state"
# Ville Bergholm 2008-2014

import unittest
from numpy import prod, sqrt, log2, kron
from numpy.random import rand, randn
from numpy.linalg import norm

# HACK to import the module in this source distribution, not the installed one!
import sys, os
sys.path.insert(0, os.path.abspath('.'))

from qit import version
from qit.base  import tol
from qit.lmap  import lmap
from qit.state import state
from qit.utils import rand_positive, rand_U, mkron


class StateConstructorTest(unittest.TestCase):
    def test_constructor(self):
        "Test state.__init__"

        # copy constructor
        temp = lmap(randn(4, 4), ((4,), (4,)))
        s = state(temp, (2, 2))
        # strings
        s = state('10011')
        s = state('2014', (3, 2, 3, 5))
        s = state('GHZ')
        s = state('GHZ', (3, 2, 3))
        s = state('W', (2, 3, 2))
        s = state('Bell2')
        # basis kets
        s = state(4, 6)
        s = state(11, (3, 5, 2))
        # kets and state ops
        s = state(rand(5))
        s = state(rand(6), (3, 2))
        s = state(rand(3, 3))
        s = state(rand(4, 4), (2, 2))
        # bad inputs
        temp = lmap(randn(2, 3), ((2,), (3,)))
        self.assertRaises(ValueError, state, temp)          # nonsquare lmap
        self.assertRaises(ValueError, state, 'rubbish')     # unknown state name
        self.assertRaises(ValueError, state, 0)             # missing dimension
        self.assertRaises(ValueError, state, 2, 2)          # ket number too high
        self.assertRaises(ValueError, state, [])            # bad array dimesion (0)
        self.assertRaises(ValueError, state, rand(2, 2, 2)) # bad array dimesion (3)
        self.assertRaises(ValueError, state, randn(3, 4))   # nonsquare array



class StateMethodTest(unittest.TestCase):
    def setUp(self):
        dim = (2, 5, 3)
        self.dim = dim
        # mixed state
        self.rho = state(rand_positive(prod(dim)), dim)
        # pure state
        self.psi = state(0, dim).u_propagate(rand_U(prod(dim)))
        # random unitary
        self.U = rand_U(prod(dim))

    def test_methods(self):
        # TODO concurrence, fix_phase, kraus_propagate, locc_convertible, lognegativity, measure,
        # negativity,

        D = prod(self.dim)
        bipartitions = [[0], [0, 2]]

        ### generalized Bloch vectors.

        temp = self.rho.bloch_vector()
        # round trip
        self.assertAlmostEqual((state.bloch_state(temp) -self.rho).norm(), 0, delta=tol)
        # correlation tensor is real
        self.assertAlmostEqual(norm(temp.imag), 0, delta=tol)
        # state purity limits the Frobenius norm
        self.assertTrue(norm(temp.flat) -sqrt(D) <= tol)
        # state normalization
        self.assertAlmostEqual(temp.flat[0], 1, delta=tol)


        ### entropy

        temp = self.rho.entropy()
        # zero for pure states
        self.assertAlmostEqual(self.psi.entropy(), 0, delta=tol)
        # nonnegative
        self.assertTrue(temp >= -tol)
        # upper limit is log2(D)
        self.assertTrue(temp <= log2(D) +tol)
        # invariant under unitary transformations
        self.assertAlmostEqual(temp, self.rho.u_propagate(self.U).entropy(), delta=tol)


        ### ptrace, ptranspose

        temp = self.rho.trace()
        for sys in bipartitions:
            rho_A = self.rho.ptrace(sys)
            # trace of partial trace equals total trace
            self.assertAlmostEqual(temp, rho_A.trace(), delta=tol)
        # partial trace over all subsystems equals total trace
        self.assertAlmostEqual(temp, self.rho.ptrace(range(self.rho.subsystems())).trace(), delta=tol)

        for sys in bipartitions:
            rho_T = self.rho.ptranspose(sys)
            # two ptransposes cancel
            self.assertAlmostEqual((self.rho -rho_T.ptranspose(sys)).norm(), 0, delta=tol)
            # ptranspose preserves trace
            self.assertAlmostEqual(temp, rho_T.trace(), delta=tol)

        ### schmidt
        return
        # FIXME svdvals causes a crash in schmidt! see if fixed in scipy 0.13.0.
        for sys in bipartitions:
            lambda1, u, v = self.psi.schmidt(sys, full=True)
            lambda2 = self.psi.schmidt(self.psi.invert_selection(sys))
            # squares of schmidt coefficients sum up to unity
            self.assertAlmostEqual(norm(lambda1), 1, delta=tol)
            # both subdivisions have identical schmidt coefficients
            self.assertAlmostEqual(norm(lambda1 -lambda2), 0, delta=tol)

            # decomposition is equal to the original matrix
            temp = 0
            for k in range(len(lambda1)):
                temp += kron(lambda1[k] * u[:, k], v[:, k])
            self.assertAlmostEqual(norm(self.psi.data.ravel() -temp), 0, delta=tol)

        # squared schmidt coefficients equal eigenvalues of partial trace
        #r = state(randn(30) + 1j*randn(30), [5, 6]).normalize()
        #x = r.schmidt([0]) ** 2  # FIXME crashes ipython!
        #temp = r.ptrace([1])
        #y, dummy = eighsort(temp.data)
        #self.assertAlmostEqual(norm(x-y), 0, delta=tol)



class StateMethod2Test(unittest.TestCase):
    def setUp(self):
        pass

    def test_methods(self):
        ### reorder

        dim = (2, 5, 1)
        A = rand(dim[0], dim[0])
        B = rand(dim[1], dim[1])
        C = rand(dim[2], dim[2])

        T1 = state(mkron(A, B, C), dim)
        T2 = T1.reorder([2, 0, 1])
        self.assertAlmostEqual(norm(mkron(C, A, B) - T2.data), 0, delta=tol)
        T2 = T1.reorder([1, 0, 2])
        self.assertAlmostEqual(norm(mkron(B, A, C) - T2.data), 0, delta=tol)
        T2 = T1.reorder([2, 1, 0])
        self.assertAlmostEqual(norm(mkron(C, B, A) - T2.data), 0, delta=tol)



class StateBinaryFuncTest(unittest.TestCase):
    def setUp(self):
        dim = (2, 3)
        # two mixed states
        self.rho1 = state(rand_positive(prod(dim)), dim)
        self.rho2 = state(rand_positive(prod(dim)), dim)
        # two pure states
        p = state(0, dim)
        self.p1 = p.u_propagate(rand_U(prod(dim)))
        self.p2 = p.u_propagate(rand_U(prod(dim)))
        # random unitary
        self.U = rand_U(prod(dim))

    def test_binary_funcs(self):
        from qit.state import fidelity, trace_dist

        fid = fidelity(self.rho1, self.rho2)
        trd = trace_dist(self.rho1, self.rho2)

        # symmetry
        self.assertAlmostEqual(fid, fidelity(self.rho2, self.rho1), delta=tol)
        self.assertAlmostEqual(trd, trace_dist(self.rho2, self.rho1), delta=tol)

        # fidelity with self, distance from self
        self.assertAlmostEqual(fidelity(self.rho1, self.rho1), 1, delta=tol)
        self.assertAlmostEqual(trace_dist(self.rho1, self.rho1), 0, delta=tol)

        # unaffected by unitary transformations
        self.assertAlmostEqual(fid, fidelity(self.rho1.u_propagate(self.U), self.rho2.u_propagate(self.U)), delta=tol)
        self.assertAlmostEqual(trd, trace_dist(self.rho1.u_propagate(self.U), self.rho2.u_propagate(self.U)), delta=tol)

        # for pure states trace_dist and fidelity are equivalent
        self.assertAlmostEqual(trace_dist(self.p1, self.p2) ** 2 +fidelity(self.p1, self.p2) ** 2, 1, delta=tol)

        # for mixed states, these inequalities hold
        self.assertTrue(sqrt(1 -fid ** 2) -trd >= -tol)
        self.assertTrue(1 -fid -trd <= tol)

        # for a pure and a mixed state we get this inequality
        self.assertTrue(1 -fidelity(self.rho1, self.p1) ** 2 -trace_dist(self.rho1, self.p1) <= tol)



if __name__ == '__main__':
    #print(sys.path)
    print('Testing QIT version ' + version())
    unittest.main()



# -*- coding: utf-8 -*-
"""
Model Hamiltonians (:mod:`qit.hamiltonian`)
===========================================

This module has methods that generate several common types of model Hamiltonians used in quantum mechanics.


.. currentmodule:: qit.hamiltonian

Contents
--------

.. autosummary::

   heisenberg
   jaynes_cummings
   hubbard
   bose_hubbard
   holstein
"""
# Ville Bergholm 2014

from __future__ import division, absolute_import, print_function, unicode_literals

import numpy as np
from numpy import asarray, conj, transpose, dot

from .base import sx, sz
from .utils import angular_momentum, op_list, boson_ladder, fermion_ladder


__all__ = [
    'heisenberg',
    'jaynes_cummings',
    'hubbard',
    'bose_hubbard',
    'holstein']


def _cdot(v, A):
    """Real dot product of a vector and a tuple of operators."""
    res = 0j
    for vv, AA in zip(v, A):
        res += vv * AA
    return res


def heisenberg(dim, C=None, J=(0, 0, 2), B=(0, 0, 1)):
    r"""Heisenberg spin network model.

    Returns the Hamiltonian H for the Heisenberg model, describing a network
    of n interacting spins in an external magnetic field.

    dim is an n-tuple of the dimensions of the spins, i.e. dim == (2, 2, 2)
    would be a system of three spin-1/2's.

    C is the :math:`n \times n` connection matrix of the spin network, where C[i,j]
    is the coupling strength between spins i and j. Only the upper triangle is used.

    J defines the form of the spin-spin interaction. It is either a 3-tuple or a
    function J(i, j) returning a 3-tuple for site-dependent interactions.
    Element k of the tuple is the coefficient of the Hamiltonian term :math:`S_k^{(i)} S_k^{(j)}`,
    where :math:`S_k^{(i)}` is the k-component of the angular momentum of spin i.

    B defines the effective magnetic field the spins locally couple to. It's either
    a 3-tuple (homogeneous field) or a function B(a) that returns a 3-tuple for
    site-dependent field.

    .. math::

      H = \sum_{\langle i,j \rangle} \sum_{k = x,y,z} J(i,j)[k] S_k^{(i)} S_k^{(j)}  +\sum_i \vec{B}(i) \cdot \vec{S}^{(i)})

    Examples::

      C = np.eye(n, n, 1)  linear n-spin chain
      J = (2, 2, 2)        isotropic Heisenberg coupling
      J = (2, 2, 0)        XX+YY coupling
      J = (0, 0, 2)        Ising ZZ coupling
      B = (0, 0, 1)        homogeneous Z-aligned field
    """
    # Ville Bergholm 2009-2014

    n = len(dim) # number of spins in the network

    if C == None:
        # linear chain
        C = np.eye(n, n, 1)

    # make J and B into functions
    if isinstance(J, tuple):
        if len(J) != 3:
            raise ValueError('J must be either a 3-tuple or a function.')
        J = asarray(J)
        Jf = lambda i, j: C[i, j] * J
    else:
        Jf = J

    if isinstance(B, tuple):
        if len(B) != 3:
            raise ValueError('B must be either a 3-tuple or a function.')
        Bf = lambda i: B
    else:
        Bf = B

    # local magnetic field terms
    temp = []
    for i in range(n):
        A = angular_momentum(dim[i])  # spin ops
        temp.append([(_cdot(Bf(i), A), i)])
    H = op_list(temp, dim)

    # spin-spin couplings: loop over nonzero entries of C
    # only use the upper triangle
    C = np.triu(C)
    for i, j in transpose(C.nonzero()):
        # spin ops for sites i and j
        Si = angular_momentum(dim[i])
        Sj = angular_momentum(dim[j])
        temp = []
        # coupling between sites a and b
        c = Jf(i, j)
        for k in range(3):
            temp.append([(c[k] * Si[k], i), (Sj[k], j)])
        H += op_list(temp, dim)

    return H, dim



def jaynes_cummings(om_a, om_c, omega, m=10):
    r"""Jaynes-Cummings model, a two-level atom in a single-mode cavity.

    Returns the Hamiltonian H and the dimension vector dim for an
    implementation of the Jaynes-Cummings model, describing a two-level atom coupled
    to a harmonic oscillator (e.g. a single EM field mode in an optical cavity).

    .. math::

      H/\hbar = \frac{\omega_a}{2} \sigma_z +\omega_c a^\dagger a +\frac{\Omega}{2} \sigma_x (a+a^\dagger)

    The dimension of the Hilbert space of the bosonic cavity mode (infinite in principle) is truncated to m.
    """
    # Ville Bergholm 2014

    dim = (2, m)
    a = boson_ladder(m) 
    ax = a.conj().transpose()

    temp = [[(om_a/2 * sz, 0)],
            [(om_c * dot(ax, a), 1)],
            [(omega * sx, 0), (a+ax, 1)]]
    H = op_list(temp, dim)

    return H, dim



def hubbard(C, U=1, mu=0):
    r"""Hubbard model, fermions on a lattice.

    Returns the Hamiltonian H and the dimension vector dim for an
    implementation of the Hubbard model.

    The model consists of spin-1/2 fermions confined in a graph defined by the
    symmetric connection matrix C (only upper triangle is used).
    The fermions interact with other fermions at the same site with interaction strength U,
    as well as with an external chemical potential mu.
    The Hamiltonian has been normalized by the fermion hopping constant t.

    .. math::

      H = -\sum_{\langle i,j \rangle, \sigma} c^\dagger_{i,\sigma} c_{j,\sigma}
        +\frac{U}{t} \sum_i n_{i,up} n_{i,down} -\frac{\mu}{t} \sum_i (n_{i,up}+n_{i,down})
    """
    # Ville Bergholm 2010-2014

    n = len(C)
    dim = 2 * np.ones(2*n)  # n sites, two fermionic modes per site

    # fermion annihilation ops f[site, spin]
    f = fermion_ladder(2 * n).reshape((n, 2))
    # NOTE all the f ops have the full Hilbert space dimension

    H = 0j

    for k in range(n):
        # number operators for this site
        n1 = dot(f[k, 0].conj().transpose(), f[k, 0])
        n2 = dot(f[k, 1].conj().transpose(), f[k, 1])
        # on-site interaction
        H += U * dot(n1, n2)
        # chemical potential
        H += -mu * (n1 + n2)

    # fermions hopping: loop over nonzero entries of C
    # only use the upper triangle
    C = np.triu(C)
    for i, j in transpose(C.nonzero()):
        for s in range(2):
            H -= dot(f[i, s].conj().transpose(), f[j, s]) +dot(f[j, s].conj().transpose(), f[i, s])

    return H, dim



def bose_hubbard(C, U=1, mu=0, m=10):
    r"""Bose-Hubbard model, bosons on a lattice.

    Returns the Hamiltonian H and the dimension vector dim for an
    implementation of the Bose-Hubbard model.

    The model consists of spinless bosons confined in a graph defined by the
    symmetric connection matrix C (only upper triangle is used).
    The bosons interact with other bosons at the same site with interaction strength U,
    as well as with an external chemical potential mu.
    The Hamiltonian has been normalized by the boson hopping constant t.

    .. math::

      H = -\sum_{\langle i,j \rangle} b^\dagger_i b_{j} +\frac{U}{2t} \sum_i n_i (n_i-1) -\frac{\mu}{t} \sum_i n_i

    The dimensions of the boson Hilbert spaces (infinite in principle) are truncated to m.
    """
    # Ville Bergholm 2010-2014

    n = len(C)
    dim = m * np.ones(n)

    b = boson_ladder(m)  # boson annihilation op
    b_dagger = b.conj().transpose()  # boson creation op
    nb = dot(b_dagger, b)  # boson number op

    I = np.eye(m)
    A = U/2 * dot(nb, (nb-I)) # on-site interaction
    B = -mu * nb # chemical potential

    temp = []
    for k in range(n):
        temp.append([(A+B, k)])
    H = op_list(temp, dim)

    temp = []
    # bosons hopping: loop over nonzero entries of C
    # only use the upper triangle
    C = np.triu(C)
    for i, j in transpose(C.nonzero()):
        temp.extend([[(b_dagger, i), (b, j)], [(b, i), (b_dagger, j)]])

    H -= op_list(temp, dim)
    return H, dim



def holstein(C, omega=1, g=1, m=10):
    r"""Holstein model, electrons on a lattice coupled to phonons.

    Returns the Hamiltonian H and the dimension vector dim for an implementation of the Holstein model.

    The model consists of spinless electrons confined in a graph defined by the
    symmetric connection matrix C (only upper triangle is used),
    coupled to phonon modes represented by a harmonic oscillator at each site.
    The dimensions of phonon Hilbert spaces (infinite in principle) are truncated to m.

    The order of the subsystems is [e1, ..., en, p1, ..., pn].
    The Hamiltonian has been normalized by the electron hopping constant t.

    .. math::

      H = -\sum_{\langle i,j \rangle} c_i^\dagger c_j  +\frac{\omega}{t} \sum_i b_i^\dagger b_i
        -\frac{g \omega}{t} \sum_i (b_i + b_i^\dagger) c_i^\dagger c_i
    """
    # Ville Bergholm 2010-2014

    n = len(C)
    # Hilbert space: electrons first, then phonons
    dim = np.r_[2**n, m * np.ones(n)]  # Jordan-Wigner clumps all fermion dims together

    c = fermion_ladder(n)  # electron annihilation ops
    b = boson_ladder(m)    # phonon annihilation
    b_dagger = b.conj().transpose()  # phonon creation
    q = b + b_dagger       # phonon position
    nb = dot(b_dagger, b)  # phonon number operator

    temp = []
    for k in range(n):
        # phonon harmonic oscillators
        temp.append([(omega * nb, 1+k)])
        # electron-phonon interaction
        temp.append([(-g * omega * dot(c[k].conj().transpose(), c[k]), 0), (q, 1+k)])
    H = op_list(temp, dim)

    # fermions hopping: loop over nonzero entries of C
    # only use the upper triangle
    C = np.triu(C)
    T = 0j
    for i, j in transpose(C.nonzero()):
        T += dot(c[i].conj().transpose(), c[j]) +dot(c[j].conj().transpose(), c[i])
    H += op_list([[(-T, 0)]], dim)

    # actual dimensions
    #dim = [2*ones(1, n), m*ones(1, n)]
    return H, dim



# -*- coding: utf-8 -*-
"Unit tests for qit.markov"
# Ville Bergholm 2009-2014

import unittest
from numpy.random import rand, randn
from numpy.linalg import norm

# HACK to import the module in this source distribution, not the installed one!
import sys, os
sys.path.insert(0, os.path.abspath('.'))

from qit import version
from qit.base import tol
from qit.utils import rand_hermitian, superop_lindblad
from qit.markov import *



class SeqTest(unittest.TestCase):
    def test_funcs(self):
        """Testing the Markovian bath module."""

        dim = 6
        H = rand_hermitian(dim)
        D = [rand_hermitian(dim)/10, rand_hermitian(dim)/10]
        B = [bath('ohmic', 1e9, 0.02), bath('ohmic', 1e9, 0.03)]

        # jump operators
        dH, X = ops(H, D)
        self.assertAlmostEqual(dH[0], 0, delta=tol)
        for n, A in enumerate(X):
            temp = A[0] # dH[0] == 0
            for k in range(1, len(dH)):
                temp += A[k] +A[k].conj().transpose() # A(-omega) == A'(omega)
            self.assertAlmostEqual(norm(temp - D[n]), 0, delta=tol) # Lindblad ops should sum to D

        # equivalence of Lindblad operators and the Liouvillian superoperator
        LL, H_LS = lindblad_ops(H, D, B)
        S1 = superop_lindblad(LL, H + H_LS)
        S2 = superop(H, D, B)
        self.assertAlmostEqual(norm(S1 - S2), 0, delta=tol)



if __name__ == '__main__':
    print('Testing QIT version ' + version())
    unittest.main()



# -*- coding: utf-8 -*-
"Unit tests for qit.ho"
# Ville Bergholm 2009-2014

import unittest
from numpy import eye, diag, ones, mat
from numpy.random import rand, randn
from numpy.linalg import norm

# HACK to import the module in this source distribution, not the installed one!
import sys, os
sys.path.insert(0, os.path.abspath('.'))

from qit import version
from qit.base import tol
from qit.state import state
from qit.ho import *
from qit.utils import boson_ladder, comm



class HOTest(unittest.TestCase):
    def setUp(self):
        pass
        

    def test_funcs(self):
        """Testing the harmonic oscillator module."""

        def randc():
            "Random complex number."
            return randn() + 1j*randn()

        n = 35
        s0 = state(0, n)

        ### displacement
        alpha = randc()
        D = displace(alpha, n=n)
        s = coherent_state(alpha, n=n)
        self.assertAlmostEqual((s -s0.u_propagate(D)).norm(), 0, delta=tol)

        ### squeeezing TODO assert
        z = randc()
        S = squeeze(z, n=n)

        ### position and momentum operators, eigenstates
        Q = position(n)
        P = momentum(n)
        q = randn()
        p = randn()
        sq = position_state(q, n=n)
        sp = momentum_state(p, n=n)
        temp = 1e-1 # the truncation accuracy is not amazing here TODO why?
        # expectation values in eigenstates
        self.assertAlmostEqual(sq.ev(Q), q, delta=temp)
        self.assertAlmostEqual(sp.ev(P), p, delta=temp)
        # [Q, P] = i
        temp = ones(n)
        temp[-1] = -n+1 # truncation...
        self.assertAlmostEqual(norm(comm(Q, P) -1j * diag(temp)), 0, delta=tol)
        # P^2 +Q^2 = 2a^\dagger * a + 1
        a = mat(boson_ladder(n))
        self.assertAlmostEqual(norm(mat(P)**2 +mat(Q)**2 -2 * a.H * a -diag(temp)), 0, delta=tol)



if __name__ == '__main__':
    print('Testing QIT version ' + version())
    unittest.main()



# -*- coding: utf-8 -*-
"Unit tests for qit.invariant"
# Ville Bergholm 2010-2014

import unittest
from numpy import kron
from numpy.linalg import norm

# HACK to import the module in this source distribution, not the installed one!
import sys, os
sys.path.insert(0, os.path.abspath('.'))

from qit import version
from qit.base import sx, tol
from qit.utils import rand_U
import qit.gate as gate
from qit.invariant import *


class InvTest(unittest.TestCase):
    def test_funcs(self):
        """Testing the invariants module."""

        U = rand_U(4)  # random two-qubit gate
        L = kron(rand_U(2), rand_U(2))  # random local 2-qubit gate
        cnot = gate.controlled(sx).data
        swap = gate.swap(2, 2).data

        # canonical invariants
        #self.assertAlmostEqual(norm(canonical(L) -[0, 0, 0]), 0, delta=tol) # only point in Berkeley chamber with translation degeneracy, (0,0,0) =^ (1,0,0)
        self.assertAlmostEqual(norm(canonical(cnot) -[0.5, 0, 0]), 0, delta=tol)
        self.assertAlmostEqual(norm(canonical(swap) -[0.5, 0.5, 0.5]), 0, delta=tol)

        # Makhlin invariants
        c = canonical(U)
        g1 = makhlin(c)
        g2 = makhlin(U)
        self.assertAlmostEqual(norm(g1-g2), 0, delta=tol)

        # maximum concurrence
        self.assertAlmostEqual(max_concurrence(L), 0, delta=tol)
        self.assertAlmostEqual(max_concurrence(swap), 0, delta=tol)
        self.assertAlmostEqual(max_concurrence(cnot), 1, delta=tol)

        #plot_weyl_2q()
        #plot_makhlin_2q(25, 25)


if __name__ == '__main__':
    print('Testing QIT version ' + version())
    unittest.main()



# -*- coding: utf-8 -*-
"""
Quantum gates (:mod:`qit.gate`)
===============================

This module implements many common types of quantum gates (and some
other useful linear maps). The returned gates are represented as sparse :class:`lmap` instances.
The data type is float unless complex entries are actually needed.

.. currentmodule:: qit.gate


Contents
--------

.. autosummary::

   dist
   id
   phase
   qft
   swap
   walsh
   mod_inc
   mod_mul
   mod_add
   controlled
   single
   two
"""

from __future__ import division, absolute_import, print_function, unicode_literals

from numpy import pi, prod, empty, zeros, trace, exp, sqrt, mod, isscalar, kron, array, ones, array_equal
import scipy.sparse as sparse

from .lmap import lmap, tensor
from .utils import qubits, op_list, assert_o, copy_memoize, gcd


__all__ = ['dist', 'id', 'phase', 'qft', 'swap', 'walsh',
           'mod_add', 'mod_inc', 'mod_mul', 
           'controlled', 'single', 'two']

# TODO reshape will cause problems for sparse matrices!
# TODO utils.op_list too!
# TODO which one is faster in element assignment -style init, dok or lil?
# TODO make input interface consistent, do we want arrays or lmaps?

def eye(D):
    """FIXME Temp. wrapper, unnecessary after when we update to scipy 0.12
    Then we can just do   from scipy.sparse import eye
    """
    return sparse.eye(D, D) 


def dist(A, B):
    r"""Distance between two unitary lmaps.

    Returns

    .. math::

       \inf_{\phi \in \mathbb{R}} \|A - e^{i \phi} B\|_F^2
       = 2 (\dim_A - |\mathrm{Tr}(A^\dagger B)|)
    """
    # Ville Bergholm 2007-2010

    if not A.is_compatible(B):
        raise ValueError('The lmaps have different dimensions.')

    temp = A.ctranspose() * B
    return 2 * (prod(temp.dim[0]) - abs(trace(temp.data)))



def id(dim):
    """Identity gate.

    Returns the identity gate I for the specified system.
    dim is a tuple of subsystem dimensions.
    """
    if isscalar(dim):
        dim = (dim,)  # scalar into a tuple
    return lmap(eye(prod(dim)), (dim, dim))


def mod_add(dim1, dim2, N=None):
    r"""Modular adder gate.

    Returns the gate U, which, operating on the computational state
    :math:`|x, y\rangle`, produces
    :math:`|x, y+x (\mod N)\rangle`.
    dim1 and dim2 are the control and target register dimension vectors.

    By default N == prod(dim2).
    If N is explicitly given we must have N <= prod(dim2), and U will act trivially on target states >= N.

    Notes:
    The modular subtractor gate can be obtained by taking the
    Hermitian conjugate of mod_add.
    mod_add(2, 2) is equal to CNOT.
    """
    # Ville Bergholm 2010

    d1 = prod(dim1)
    d2 = prod(dim2)
    if N == None:
        N = d2
    elif d2 < N:
        raise ValueError('Target register dimension must be >= N.')

    # NOTE: a real quantum computer would implement this gate using a
    # sequence of reversible arithmetic gates but since we don't have
    # one we might as well cheat
    dim = d1 * d2
    U = sparse.dok_matrix((dim, dim))
    for a in range(d1):
        for b in range(d2):
            y = d2*a + b
            if b < N:
                x = d2*a +mod(a+b, N)
            else:
                # U acts trivially for target states >= N
                x = y
            U[x, y] = 1

    dim = (dim1, dim2)
    return lmap(U.tocsr(), (dim, dim))


def mod_inc(x, dim, N=None):
    r"""Modular incrementation gate.

    U = mod_inc(x, dim)     N == prod(dim)
    U = mod_inc(x, dim, N)  gate dimension prod(dim) must be >= N

    Returns the gate U, which, operating on the computational state
    :math:`|y\rangle`, increments it by x (mod N):
    :math:`U |y\rangle = |y+x (mod N)\rangle`.

    If N is given, U will act trivially on computational states >= N.
    """
    # Ville Bergholm 2010

    if isscalar(dim):
        dim = (dim,)  # scalar into a tuple
    d = prod(dim)
    if N == None:
        N = d
    elif d < N:
        raise ValueError('Gate dimension must be >= N.')

    U = sparse.dok_matrix((d, d))
    for y in range(N):
        U[mod(x+y, N), y] = 1
    # U acts trivially for states >= N
    for y in range(N, d):
        U[y, y] = 1

    return lmap(U.tocsr(), (dim, dim))


def mod_mul(x, dim, N=None):
    r"""Modular multiplication gate.

    U = mod_mul(x, dim)     N == prod(dim)
    U = mod_mul(x, dim, N)  gate dimension prod(dim) must be >= N

    Returns the gate U, which, operating on the computational state
    :math:`|y\rangle`, multiplies it by x (mod N):
    :math:`U |y\rangle = |x*y (mod N)\rangle`.
    x and N must be coprime for the operation to be reversible.

    If N is given, U will act trivially on computational states >= N.
    """
    # Ville Bergholm 2010-2011

    if isscalar(dim):
        dim = (dim,)  # scalar into a tuple
    d = prod(dim)
    if N == None:
        N = d
    elif d < N:
        raise ValueError('Gate dimension must be >= N.')

    if gcd(x, N) != 1:
        raise ValueError('x and N must be coprime for the mul operation to be reversible.')

    # NOTE: a real quantum computer would implement this gate using a
    # sequence of reversible arithmetic gates but since we don't have
    # one we might as well cheat
    U = sparse.dok_matrix((d, d))
    for y in range(N):
        U[mod(x*y, N), y] = 1
    # U acts trivially for states >= N
    for y in range(N, d):
        U[y, y] = 1

    return lmap(U.tocsr(), (dim, dim))


def phase(theta, dim=None):
    """Diagonal phase shift gate.

    Returns the (diagonal) phase shift gate U = diag(exp(i*theta)).
    """
    # Ville Bergholm 2011

    if isscalar(dim):
        dim = (dim,)  # scalar into a tuple
    n = len(theta)
    if dim == None:
        dim = (n,)
    d = prod(dim)
    if d != n:
        raise ValueError('Dimension mismatch.')

    return lmap(sparse.diags(exp(1j * theta), 0) , (dim, dim))


@copy_memoize
def qft(dim):
    """Quantum Fourier transform gate.

    Returns the quantum Fourier transform gate for the specified system.
    dim is a vector of subsystem dimensions.
    The returned lmap is dense.
    """
    # Ville Bergholm 2004-2011

    if isscalar(dim):
        dim = (dim,)  # scalar into a tuple
    N = prod(dim)
    U = empty((N, N), complex)  # completely dense, so we don't have to initialize it with zeros
    for j in range(N):
        for k in range(N):
            U[j, k] = exp(2j * pi * j * k / N) / sqrt(N)
    return lmap(U, (dim, dim))


def swap(d1, d2):
    r"""Swap gate.

    Returns the swap gate which swaps the order of two subsystems with dimensions [d1, d2].

    .. math::

       S: A_1 \otimes A_2 \to A_2 \otimes A_1, \quad
       v_1 \otimes v_2 \mapsto v_2 \otimes v_1
    """
    # Ville Bergholm 2010

    temp = d1*d2
    U = sparse.dok_matrix((temp, temp))
    for x in range(d1):
        for y in range(d2):
            U[d1*y + x, d2*x + y] = 1
    return lmap(U.tocsr(), ((d2, d1), (d1, d2)))


def walsh(n):
    """Walsh-Hadamard gate.

    Returns the Walsh-Hadamard gate for n qubits.
    The returned lmap is dense.
    """
    # Ville Bergholm 2009-2010

    from .base import H

    U = 1
    for _ in range(n):
        U = kron(U, H)
    dim = qubits(n)
    return lmap(U, (dim, dim))


def controlled(U, ctrl=(1,), dim=None):
    r"""Controlled gate.

    Returns the (t+1)-qudit controlled-U gate, where t == length(ctrl).
    U has to be a square matrix.

    ctrl is an integer vector defining the control nodes. It has one entry k per
    control qudit, denoting the required computational basis state :math:`|k\rangle`
    for that particular qudit. Value k == -1 denotes no control.

    dim is the dimensions vector for the control qudits. If not given, all controls
    are assumed to be qubits.

    Examples::

      controlled(NOT, [1]) gives the standard CNOT gate.
      controlled(NOT, [1, 1]) gives the Toffoli gate.
    """
    # Ville Bergholm 2009-2011

    # TODO generalization, uniformly controlled gates?
    if isscalar(dim):
        dim = (dim,)  # scalar into a tuple
    t = len(ctrl)
    if dim == None:
        dim = qubits(t) # qubits by default

    if t != len(dim):
        raise ValueError('ctrl and dim vectors have unequal lengths.')

    if any(array(ctrl) >= array(dim)):
        raise ValueError('Control on non-existant state.')

    yes = 1  # just the diagonal
    for k in range(t):
        if ctrl[k] >= 0:
            temp = zeros(dim[k])
            temp[ctrl[k]] = 1  # control on k
            yes = kron(yes, temp) 
        else:
            yes = kron(yes, ones(dim[k])) # no control on this qudit

    no = 1 - yes
    T = prod(dim)
    dim = list(dim)

    if isinstance(U, lmap):
        d1 = dim + list(U.dim[0])
        d2 = dim + list(U.dim[1])
        U = U.data
    else:
        d1 = dim + [U.shape[0]]
        d2 = dim + [U.shape[1]]

    # controlled gates only make sense for square matrices U (we need an identity transformation for the 'no' cases!)
    U_dim = U.shape[0]
    out = sparse.diags(kron(no, ones(U_dim)), 0) +sparse.kron(sparse.diags(yes, 0), U)
    return lmap(out, (d1, d2))


def single(L, t, d_in):
    """Single-qudit operator.

    Returns the operator U corresponding to the local operator L applied
    to subsystem t (and identity applied to the remaining subsystems).

    d_in is the input dimension vector for U.
    """
    # James Whitfield 2010
    # Ville Bergholm 2010

    if isinstance(L, lmap):
        L = L.data  # into ndarray

    d_in = list(d_in)
    if d_in[t] != L.shape[1]:
        raise ValueError('Input dimensions do not match.')
    d_out = d_in
    d_out[t] = L.shape[0]
    return lmap(op_list([[[L, t]]], d_in), (d_out, d_in))


def two(B, t, d_in):
    """Two-qudit operator.

    Returns the operator U corresponding to the bipartite operator B applied
    to subsystems t == [t1, t2] (and identity applied to the remaining subsystems).

    d_in is the input dimension vector for U.
    """
    # James Whitfield 2010
    # Ville Bergholm 2010-2011

    if len(t) != 2:
        raise ValueError('Exactly two target subsystems required.')

    n = len(d_in)
    t = array(t)
    if any(t < 0) or any(t >= n) or t[0] == t[1]:
        raise ValueError('Bad target subsystem(s).')

    d_in = array(d_in)
    if not array_equal(d_in[t], B.dim[1]):
        raise ValueError('Input dimensions do not match.')

    # dimensions for the untouched subsystems
    a = min(t)
    b = max(t)
    before    = prod(d_in[:a])
    inbetween = prod(d_in[a+1:b])
    after     = prod(d_in[b+1:])

    # how tensor(B_{01}, I_2) should be reordered
    if t[0] < t[1]:
        p = [0, 2, 1]
    else:
        p = [1, 2, 0]
    U = tensor(B, lmap(eye(inbetween))).reorder((p, p), inplace = True)
    U = tensor(lmap(eye(before)), U, lmap(eye(after)))

    # restore dimensions
    d_out = d_in.copy()
    d_out[t] = B.dim[0]
    return lmap(U, (d_out, d_in))
